import tensorflow as tf
import numpy as np
from dryml.utils import get_temp_checkpoint_dir, cleanup_checkpoint_dir, \
    adjust_class_module
from dryml import Meta
from dryml.data import Dataset
from dryml.data.tf import TFDataset
from dryml.models.trainable import Trainable as BaseTrainable
from dryml.models.tf.base import Model as TFModel
from dryml.models.tf.base import Trainable as TFTrainable
from dryml.models.tf.base import TrainFunction as TFTrainFunction
from dryml.models.tf.base import Wrapper
from dryml.models.tf.utils import keras_train_spec_updater, \
    keras_callback_wrapper
import tempfile
import zipfile
import os
import re
import shutil
from typing import List


def keras_load_checkpoint_from_zip(
        mdl: tf.keras.Model,
        zf: zipfile.ZipFile,
        checkpoint_name: str,
        checkpoint_dir: str = 'checkpoints') -> bool:
    if mdl is None:
        raise RuntimeError(
            "keras model can't be None! "
            "Did you run the object's compute_prepare method?")

    try:
        # Get namelist
        ns = zf.namelist()

        # Detect actual checkpoint name
        checkpoint_name_re = re.compile(f"({checkpoint_name}.*)\\.index")

        checkpoint_name_matches = []
        for n in ns:
            re_res = checkpoint_name_re.search(n)
            if re_res is not None:
                checkpoint_name_matches.append(re_res.groups(1)[0])

        if len(checkpoint_name_matches) > 1:
            print("Warning! More than one checkpoing matched in save file!")

        checkpoint_name = checkpoint_name_matches[-1]

        # Add a slash at the end
        if checkpoint_dir[-1] != '/':
            checkpoint_dir += '/'

        # Get files inside the checkpoint directory
        orig_ns = list(filter(lambda n: n.startswith(checkpoint_dir), ns))
        if len(orig_ns) == 0:
            raise RuntimeError(f"No directory {checkpoint_dir} in zipfile!")

        # Get destination names
        dest_ns = list(map(lambda n: n[len(checkpoint_dir):], orig_ns))

        ns = zip(orig_ns, dest_ns)

        checkpoint_file_re = re.compile(f"{checkpoint_name}.*\\.(index|data)")
        ns = filter(lambda t: checkpoint_file_re.search(t[0]) is not None, ns)

        # Create temp directory
        with tempfile.TemporaryDirectory() as d:
            # Extract files from zipfile to the temp directory
            for orig_n, dest_n in ns:
                zf.extract(orig_n, path=d)

            # Load the weights into the model with the load weights function
            checkpoint_path = os.path.join(d, checkpoint_dir, checkpoint_name)
            # expect to load partial because we may
            # not have optimizer in place yet.
            tf.train.Checkpoint(mdl).restore(checkpoint_path).expect_partial()
    except Exception as e:
        print(f"Issue loading checkpoint! {e}")
        return False
    return True


def keras_load_checkpoint_from_zip_to_dir(
        mdl: tf.keras.Model,
        zf: zipfile.ZipFile,
        checkpoint_name: str,
        temp_checkpoint_dir: str,
        zip_checkpoint_dir: str = 'checkpoints') -> bool:
    if mdl is None:
        raise ValueError(
            "keras model can't be None! "
            "Did you run the object's compute_prepare method?")

    try:
        # Get namelist
        ns = zf.namelist()

        # Detect actual checkpoint name
        checkpoint_name_re = re.compile(f"({checkpoint_name}.*)\\.index")

        checkpoint_name_matches = []
        for n in ns:
            re_res = checkpoint_name_re.search(n)
            if re_res is not None:
                checkpoint_name_matches.append(re_res.groups(1)[0])

        if len(checkpoint_name_matches) > 1:
            print("Warning! More than one checkpoing matched in save file!")

        checkpoint_name = checkpoint_name_matches[-1]

        # Add a slash at the end
        if zip_checkpoint_dir[-1] != '/':
            zip_checkpoint_dir += '/'

        # Get files inside the checkpoint directory
        orig_ns = list(filter(lambda n: n.startswith(zip_checkpoint_dir), ns))
        if len(orig_ns) == 0:
            raise RuntimeError(
                f"No directory {zip_checkpoint_dir} in zipfile!")

        # Get destination names
        dest_ns = list(map(lambda n: n[len(zip_checkpoint_dir):], orig_ns))

        ns = zip(orig_ns, dest_ns)

        checkpoint_file_re = re.compile(f"{checkpoint_name}.*\\.(index|data)")
        ns = filter(lambda t: checkpoint_file_re.search(t[0]) is not None, ns)

        # Extract files from zipfile to the temp directory
        for orig_n, dest_n in ns:
            zf.extract(orig_n, path=temp_checkpoint_dir)

        # Load the weights into the model with the load weights function
        checkpoint_path = os.path.join(
            temp_checkpoint_dir, 'checkpoints', checkpoint_name)

        # expect to load partial because we may
        # not have optimizer in place yet.
        tf.train.Checkpoint(model=mdl) \
                .restore(checkpoint_path) \
                .assert_existing_objects_matched() \
                .expect_partial()
    except Exception as e:
        print(f"Issue loading checkpoint! {e}")
        return False
    return True


def keras_save_checkpoint_to_zip(
        mdl: tf.keras.Model,
        zf: zipfile.ZipFile,
        checkpoint_name: str,
        checkpoint_dir: str = 'checkpoints') -> bool:
    try:
        # Adjust checkpoint directory
        if checkpoint_dir[-1] != '/':
            checkpoint_dir += '/'

        # Get namelist
        ns = zf.namelist()

        # Get files inside the checkpoint directory
        ns = filter(lambda n: n.startswith(checkpoint_dir), ns)
        checkpoint_file_re = re.compile(f"{checkpoint_name}\\.(index|data)")
        ns = list(filter(
            lambda n: checkpoint_file_re.search(n) is not None,
            ns))

        # Save checkpoint to a temporary directory
        with tempfile.TemporaryDirectory() as d:
            # Load the weights into the model with the load weights function
            checkpoint_path = os.path.join(d, checkpoint_name)

            # Create checkpoint object and save
            checkpoint = tf.train.Checkpoint(mdl)
            checkpoint.save(checkpoint_path)

            checkpoint_files = os.listdir(d)

            # We would like to delete existing files from the directory
            # But we currently can't do this with python.
            # Instead, we should fail if the files which were written
            # for the new checkpoint don't match those in the
            # zipfile already.
            if len(ns) != 0:
                # List
                for f_name in checkpoint_files:
                    if f_name not in ns:
                        raise RuntimeError(
                            f"Zipfile {zf.filename} already has a checkpoint "
                            f"of name {checkpoint_name} saved. The new "
                            f"checkpoint file {f_name} not already in "
                            "existing checkpoint file set! (python can't "
                            "remove files from open zipfiles)")

            # Write or Overwrite files in the zipfile
            for f_name in checkpoint_files:
                zf.write(
                    os.path.join(d, f_name),
                    arcname=os.path.join(checkpoint_dir, f_name))
    except Exception as e:
        print(f"Error saving keras weights! {e}")
        return False
    return True


def keras_save_checkpoint_to_zip_from_dir(
        mdl: tf.keras.Model,
        zf: zipfile.ZipFile,
        checkpoint_name: str,
        temp_checkpoint_dir: str,
        zip_checkpoint_dir: str = 'checkpoints') -> bool:
    if mdl is None:
        raise ValueError(
            "keras model can't be None! "
            "Did you run the object's compute_prepare method?")
    try:
        # Adjust checkpoint directory
        if zip_checkpoint_dir[-1] != '/':
            zip_checkpoint_dir += '/'

        # Get namelist
        ns = zf.namelist()

        # Get files inside the checkpoint directory
        ns = filter(lambda n: n.startswith(zip_checkpoint_dir), ns)
        checkpoint_file_re = re.compile(f"{checkpoint_name}\\.(index|data)")
        ns = list(filter(
            lambda n: checkpoint_file_re.search(n) is not None,
            ns))

        # clear existing checkpoint data
        for el_name in os.listdir(temp_checkpoint_dir):
            el_path = os.path.join(temp_checkpoint_dir, el_name)
            if os.path.isdir(el_path):
                shutil.rmtree(el_path)
            else:
                os.remove(el_path)

        # get the full checkpoint path
        checkpoint_path = os.path.join(temp_checkpoint_dir, checkpoint_name)

        # Create checkpoint object and save
        checkpoint = tf.train.Checkpoint(model=mdl)
        checkpoint.save(checkpoint_path)

        # Get list of saved files
        checkpoint_files = os.listdir(temp_checkpoint_dir)

        # We would like to delete existing files from the directory
        # But we currently can't do this with python.
        # Instead, we should fail if the files which were written
        # for the new checkpoint don't match those in the
        # zipfile already.
        if len(ns) != 0:
            # List
            for f_name in checkpoint_files:
                if f_name not in ns:
                    raise RuntimeError(
                        f"Zipfile {zf.filename} already has a checkpoint "
                        f"of name {checkpoint_name} saved. The new "
                        f"checkpoint file {f_name} not already in "
                        "existing checkpoint file set! (python can't "
                        "remove files from open zipfiles)")

        # Write or Overwrite files in the zipfile
        for f_name in checkpoint_files:
            zf.write(
                os.path.join(temp_checkpoint_dir, f_name),
                arcname=os.path.join(zip_checkpoint_dir, f_name))
    except Exception as e:
        print(f"Error saving keras weights! {e}")
        return False
    return True


class Model(TFModel):
    def __init__(
            self, *args, **kwargs):
        # It is subclass's responsibility to fill this
        # attribute with an actual keras class
        self.mdl = None

    def __call__(self, X, *args, target=True, index=False, **kwargs):
        return self.mdl(X, *args, **kwargs)

    def compute_cleanup_imp(self):
        # Delete the contained model
        del self.mdl
        self.mdl = None


class ModelWrapper(Model):
    @Meta.collect_args
    @Meta.collect_kwargs
    def __init__(self, cls, *args, **kwargs):
        if type(cls) is not type:
            raise TypeError(
                f"Expected first argument to be type. Got {type(cls)}")
        self.cls = cls
        self.args = args
        self.kwargs = kwargs

    def compute_prepare_imp(self):
        self.mdl = self.cls(*self.args, **self.kwargs)


class TrainFunction(TFTrainFunction):
    pass


class BasicTraining(TrainFunction):
    def __init__(
            self, *args, val_split=0.2, val_num=None, num_total=None,
            shuffle_buffer=None, epochs=10000, **kwargs):
        self.val_split = val_split
        self.val_num = val_num
        self.shuffle_buffer = shuffle_buffer
        self.epochs = epochs
        self.num_total = num_total

    def __call__(
            self, trainable, data: Dataset, train_spec=None,
            train_callbacks=[]):

        # Pop the epoch to resume from
        start_epoch = 0
        if train_spec is not None:
            start_epoch = train_spec.level_step()

        # Type checking training data, and converting if necessary
        data = data.tf()

        # Check data is supervised.
        if not data.supervised:
            raise RuntimeError(
                "TFBasicTraining requires supervised data")

        # Make sure data is unbatched. We want the function to control this.
        data = data.unbatch()

        num_total = self.num_total
        if num_total is None:
            # Attempt to measure dataset size
            num_total = data.count()
        if self.val_num is not None:
            num_val = self.val_num
            if self.val_split is not None:
                print("Overridding val_split with val_num.")
            if self.val_num > num_total:
                raise ValueError("val_num cannot be larger than num_total!")
        else:
            num_val = int(num_total*self.val_split)

        num_train = num_total-num_val

        # Get tf.data.Dataset backing type
        ds = data.data()

        if self.shuffle_buffer is None:
            shuffle_buffer = num_train
        else:
            shuffle_buffer = self.shuffle_buffer

        batch_size = self.train_kwargs.pop('batch_size', 32)
        callbacks = self.train_kwargs.pop('callbacks', None)

        ds_val = ds.take(num_val)
        ds_val = ds_val.batch(batch_size)
        ds_val = ds_val.prefetch(tf.data.AUTOTUNE)

        ds_train = ds.skip(num_val)
        ds_train = ds_train.cache()
        ds_train = ds_train.shuffle(shuffle_buffer)
        ds_train = ds_train.batch(batch_size)
        ds_train = ds_train.prefetch(tf.data.AUTOTUNE)

        if callbacks is None:
            callbacks = []

        if train_spec is not None:
            callbacks.append(
                keras_train_spec_updater(train_spec))

        for callback in train_callbacks:
            new_callback = keras_callback_wrapper(callback)
            callbacks.append(new_callback)

        # Fit model
        trainable.model.mdl.fit(
            ds_train, *self.train_args, validation_data=ds_val,
            callbacks=callbacks, epochs=self.epochs,
            initial_epoch=start_epoch, **self.train_kwargs)


class BasicEarlyStoppingTraining(TrainFunction):
    def __init__(
            self, *args, val_split=0.2, val_num=None, num_total=None,
            shuffle_buffer=None, patience=3, epochs=10000, **kwargs):
        self.val_split = val_split
        self.val_num = val_num
        self.shuffle_buffer = shuffle_buffer
        self.patience = patience
        self.epochs = epochs
        self.num_total = num_total

    def __call__(
            self, trainable, data, train_spec=None,
            train_callbacks=[]):

        # Pop the epoch to resume from
        start_epoch = 0
        if train_spec is not None:
            start_epoch = train_spec.level_step()

        # Type checking training data, and converting if necessary
        if type(data) is not TFDataset:
            if not hasattr(data, 'to_TFDataset'):
                raise TypeError(
                    f"Type {type(data)} can't be converted to TFDataset!")
            data = data.to_TFDataset()

        # Check data is supervised.
        if not data.supervised:
            raise RuntimeError(
                "TFBasicEarlyStoppingTraining requires supervised data")

        batch_size = self.train_kwargs.pop('batch_size', 32)
        callbacks = self.train_kwargs.pop('callbacks', None)

        # Make sure data is unbatched. We want the function to control this.
        data = data.unbatch()

        num_total = self.num_total
        if num_total is None:
            # Attempt to measure dataset size
            num_total = data.count()
        if self.val_num is not None:
            num_val = self.val_num
            if self.val_split is not None:
                print("Overridding val_split with val_num.")
            if self.val_num > num_total:
                raise ValueError("val_num cannot be larger than num_total!")
        else:
            num_val = int(num_total*self.val_split)

        num_train = num_total-num_val

        # Get tf.data.Dataset backing type
        ds = data.data()

        if self.shuffle_buffer is None:
            shuffle_buffer = num_train
        else:
            shuffle_buffer = self.shuffle_buffer

        ds_val = ds.take(num_val)
        ds_val = ds_val.batch(batch_size)
        ds_val = ds_val.prefetch(tf.data.AUTOTUNE)

        ds_train = ds.skip(num_val)
        ds_train = ds_train.cache()
        ds_train = ds_train.shuffle(shuffle_buffer)
        ds_train = ds_train.batch(batch_size)
        ds_train = ds_train.prefetch(tf.data.AUTOTUNE)

        if callbacks is None:
            callbacks = []

        # Add Early Stopping Callback
        callbacks.append(tf.keras.callbacks.EarlyStopping(
            patience=self.patience,
            monitor='val_loss',
            restore_best_weights=True))

        if train_spec is not None:
            callbacks.append(
                keras_train_spec_updater(train_spec))

        for callback in train_callbacks:
            new_callback = keras_callback_wrapper(callback)
            callbacks.append(new_callback)

        # Fit model
        trainable.model.mdl.fit(
            ds_train, validation_data=ds_val,
            initial_epoch=start_epoch,
            epochs=self.epochs, callbacks=callbacks,
            *self.train_args, **self.train_kwargs)


class SequentialFunctionalModel(Model):
    def __init__(
            self, input_shape=(1,), layer_defs=[]):

        self.input_shape = input_shape
        self.layer_defs = layer_defs

    def compute_prepare_imp(self):
        # Build Functional Model
        inp = tf.keras.layers.Input(self.input_shape)
        last_layer = inp
        for layer_name, layer_kwargs in self.layer_defs:
            last_layer = getattr(
                tf.keras.layers, layer_name)(**layer_kwargs)(last_layer)
        self.mdl = tf.keras.Model(inputs=inp, outputs=last_layer)


def keras_sequential_functional_class(
        name, input_shape, output_shape, base_classes=(Model,)):

    def __init__(
            self, layer_defs, *args, out_activation='linear',
            **kwargs):
        self.layer_defs = layer_defs
        self.out_activation = out_activation

    def compute_prepare_imp(self):
        # Build Functional Model
        inp = tf.keras.layers.Input(input_shape)
        last_layer = inp
        for layer_name, layer_kwargs in self.layer_defs:
            last_layer = getattr(
                tf.keras.layers, layer_name)(**layer_kwargs)(last_layer)
        # Initially flatten result
        last_layer = tf.keras.layers.Flatten()(last_layer)
        # Compute number of output units
        output_units = np.cumprod(output_shape)
        last_layer = tf.keras.layers.Dense(
            output_units, activation=self.out_activation)(last_layer)
        # Respect final shape
        last_layer = tf.keras.layers.Reshape(output_shape)(last_layer)
        self.mdl = tf.keras.Model(inputs=inp, outputs=last_layer)

    # Create the new class
    new_cls = type(name, base_classes, {
        '__init__': __init__,
        'compute_prepare_imp': compute_prepare_imp,
    })

    adjust_class_module(new_cls)

    return new_cls


class Trainable(TFTrainable):
    def __init__(
            self, model: Model = None,
            optimizer: Wrapper = None,
            loss: Wrapper = None,
            metrics: List[Wrapper] = [],
            train_fn: TrainFunction = None):
        if model is None:
            raise ValueError(
                "You need to set the model component of this trainable!")
        self.model = model

        if optimizer is None:
            raise ValueError(
                "You need to set the optimizer component of this trainable!")
        self.optimizer = optimizer

        if loss is None:
            raise ValueError(
                "You need to set the loss component of this trainable!")
        self.loss = loss

        self.metrics = metrics

        if train_fn is None:
            raise ValueError(
                "You need to set the train_fn component of this trainable!")
        self.train_fn = train_fn

        self._temp_checkpoint_dir = None

    def compute_prepare_imp(self):
        self._temp_checkpoint_dir = get_temp_checkpoint_dir(self.dry_id)

    def compute_cleanup_imp(self):
        if self._temp_checkpoint_dir is not None:
            cleanup_checkpoint_dir(self._temp_checkpoint_dir)
            self._temp_checkpoint_dir = None

    def load_compute_imp(self, file: zipfile.ZipFile) -> bool:
        # Load Weights
        if not keras_load_checkpoint_from_zip_to_dir(
                self.model.mdl, file, 'ckpt-1', self._temp_checkpoint_dir):
            print("Error loading keras weights")
            return False

        return True

    def save_compute_imp(self, file: zipfile.ZipFile) -> bool:

        # Save Weights
        if not keras_save_checkpoint_to_zip_from_dir(
                self.model.mdl, file, 'ckpt', self._temp_checkpoint_dir):
            print("Error saving keras weights")
            return False

        return True

    def train(
            self, data, train_spec=None, train_callbacks=[],
            metrics=[]):
        self.train_fn(
            self, data, train_spec=train_spec,
            train_callbacks=train_callbacks)
        self.train_state = BaseTrainable.trained

    def prep_train(self):
        metric_list = []
        for metric in self.metrics:
            metric_list.append(metric.obj)

        self.model.mdl.compile(
            optimizer=self.optimizer.obj,
            loss=self.loss.obj,
            metrics=metric_list)

    def prep_eval(self):
        pass



"""
A module for tracking the currently available computing context
"""


from contextlib import contextmanager
from typing import Type, Union, Optional
from dryml.utils import is_nonstring_iterable
from collections import UserDict
import multiprocessing
import re


specific_resource_re = re.compile('^(cpu|gpu)/([0-9]+)$')


class InsufficientResourcesError(Exception):
    pass


class ResourceRequest(UserDict):
    """
    A request of specific resources.
    """

    def __init__(
            self, *args, **res_map):
        super().__init__()
        for arg in args:
            if type(arg) is not dict:
                raise TypeError("Non dict arguments not supported")
            self.data.update(**arg)
        self.data.update(**res_map)
        if len(self.data.keys()):
            # Nothing passed, by default only get one cpu.
            self['num_cpus'] = 1

    def __setitem__(self, key, val):
        if type(val) not in [float, int]:
            raise ValueError(
                "Can't set a non-float or non-int resource fraction")
        if key in ['num_gpus', 'num_cpus']:
            self.data[key] = val
        else:
            spec_resource_match = specific_resource_re.match(key)
            if not spec_resource_match:
                raise KeyError(f"Resource key {key} not valid.")
            self.data[key] = val


def combine_requests(requests: [ResourceRequest]):
    # In the future I might expand functionality
    # here to do combinations other than max.
    all_keys = []
    for req in requests:
        all_keys += list(req.keys())

    result_request = ResourceRequest()

    for key in all_keys:
        max_val = max(map(
            lambda r: r[key],
            filter(lambda r: key in r, requests)))
        result_request[key] = max_val

    return result_request


def combine_reqs(*requests):
    result_request = {}
    for req in requests:
        for framework in req:
            if framework not in result_request:
                result_request[framework] = [req[framework]]
            else:
                result_request[framework].append(req[framework])

    for framework in result_request:
        result_request[framework] = combine_requests(result_request[framework])

    return result_request


def resource_request_builder(req_dict):
    return ResourceRequest(req_dict)


class ResourceAllocation(UserDict):
    def __init__(
            self, **resource_map):
        super().__init__()
        self.data.update(**resource_map)

    @property
    def gpus(self):
        return list(filter(lambda k: 'gpu' in k, self.keys()))

    @property
    def num_gpus(self):
        return len(self.gpus)

    @property
    def cpus(self):
        return list(filter(lambda k: 'cpu' in k, self.keys()))

    @property
    def num_cpus(self):
        return len(self.cpus)

    def satisfies(self, request: Optional[Union[ResourceRequest, dict]] = {}):
        # Handle construction of the resource request object
        if type(request) is dict:
            request = ResourceRequest(request)

        # Get list of specific request keys
        specific_keys = list(filter(
            lambda k: k not in ['num_cpus', 'num_gpus'],
            request.keys()))
        for key in specific_keys:
            if self[key] < request[key]:
                return False

        # Get remaining resource keys
        remaining_resource_keys = list(filter(
            lambda k: k not in specific_keys,
            self.keys()))

        # Check if we have enough to satisfy num_gpu/num_cpus
        if 'num_gpus' in request:
            if request['num_gpus'] == -1:
                # All remaining gpus are requested. for now,
                # don't apply any constraint
                pass
            else:
                num_gpu_keys = len(list(filter(
                    lambda k: 'gpu' in k,
                    remaining_resource_keys)))
                if num_gpu_keys < request['num_gpus']:
                    return False
        if 'num_cpus' in request:
            if request['num_cpus'] == -1:
                # All remaining cpus are requested. for now,
                # don't apply any constraint
                pass
            else:
                num_cpu_keys = len(list(filter(
                    lambda k: 'cpu' in k,
                    remaining_resource_keys)))
                if num_cpu_keys < request['num_cpus']:
                    return False

        return True


class ResourcePool(object):
    """
    For now, the cpu entries of a resource pool are essentially dummy
    variables. They aren't really doing anything. This can be expanded
    later to handle parallelized algorithms

    Perhaps for now, 'num_gpus' should indicate how many 'full' gpus
    """
    def __init__(self, num_cpus=None, num_gpus=None, _test=False):
        self.resource_map = {}
        if not _test:
            total_num_cpus = len(list(range(multiprocessing.cpu_count())))
            if num_cpus is None:
                num_cpus = total_num_cpus
            elif num_cpus > total_num_cpus:
                raise InsufficientResourcesError("There are not enough cpus!")
        else:
            if num_cpus is None:
                raise ValueError(
                    "In testing, you must pass a value to 'num_cpus'.")

        for i in range(num_cpus):
            self.resource_map[f"cpu/{i}"] = 1.

        if not _test:
            import GPUtil
            try:
                total_num_gpus = len(GPUtil.getGPUs())
            except ValueError:
                # There was an issue retrieving gpus!
                total_num_gpus = 0
            if num_gpus is None:
                num_gpus = total_num_gpus
            elif num_gpus > total_num_gpus:
                raise InsufficientResourcesError("There are not enough gpus!")
        else:
            if num_gpus is None:
                raise ValueError(
                    "In testing, you must pass a value to 'num_gpus'.")

        for i in range(num_gpus):
            self.resource_map[f"gpu/{i}"] = 1.

    def __repr__(self):
        return self.resource_map.__repr__()

    def __str__(self):
        return self.resource_map.__str__()

    @property
    def cpus(self):
        return list(filter(lambda k: 'cpu' in k, self.resource_map.keys()))

    @property
    def num_cpus(self):
        return len(self.cpus)

    @property
    def gpus(self):
        return list(filter(lambda k: 'gpu' in k, self.resource_map.keys()))

    @property
    def num_gpus(self):
        return len(self.gpus)

    def request(self, resource_request: Union[ResourceRequest, dict]):
        # Handle construction of the resource request object
        if type(resource_request) is dict:
            resource_request = ResourceRequest(resource_request)

        alloc = ResourceAllocation()
        # First, we should allocate the specific resource requests
        for key in resource_request:
            spec_match = specific_resource_re.match(key)
            if spec_match is not None:
                avail_val = self.resource_map[key]
                ask_val = resource_request[key]
                if ask_val > avail_val:
                    self.release(alloc)
                    raise InsufficientResourcesError(
                        f"{key} doesn't have enough availability")
                alloc[key] = ask_val
                self.resource_map[key] = avail_val-ask_val

        if 'num_gpus' in resource_request and \
                resource_request['num_gpus'] != -1:
            gpu_keys = list(filter(
                lambda s: 'gpu' in s,
                self.resource_map.keys()))
            gpu_keys = sorted(
                gpu_keys,
                key=lambda k: self.resource_map[k],
                reverse=True)
            for i in range(resource_request['num_gpus']):
                gpu_key = gpu_keys[i]
                if self.resource_map[gpu_key] < 1.:
                    self.release(alloc)
                    raise InsufficientResourcesError(
                        "Couldn't find a gpu with full availability!")
                alloc[gpu_key] = 1.
                self.resource_map[gpu_key] = 0.
        else:
            if 'num_gpus' in resource_request:
                if resource_request['num_gpus'] == -1:
                    # Allocate all remaining gpus that aren't yet allocated.
                    already_allocated_gpu_keys = alloc.gpus
                    available_gpu_keys = self.gpus
                    remaining_gpu_keys = list(filter(
                        lambda k: k not in already_allocated_gpu_keys,
                        available_gpu_keys))
                    remaining_gpu_keys = list(filter(
                        lambda k: self.resource_map[k] == 1.,
                        remaining_gpu_keys))
                    for gpu_key in remaining_gpu_keys:
                        alloc[gpu_key] = 1.
                        self.resource_map[gpu_key] = 0.
                else:
                    # Report error
                    self.release(alloc)
                    raise ValueError(
                        f"'num_gpus' value {resource_request['num_gpus']}"
                        "not supported.")

        # Next, allocate general resource requests
        if 'num_cpus' in resource_request and \
                resource_request['num_cpus'] != -1:
            cpu_keys = list(filter(
                lambda s: 'cpu' in s,
                self.resource_map.keys()))
            cpu_keys = sorted(
                cpu_keys,
                key=lambda k: self.resource_map[k],
                reverse=True)
            for i in range(resource_request['num_cpus']):
                cpu_key = cpu_keys[i]
                if self.resource_map[cpu_key] < 1.:
                    self.release(alloc)
                    raise InsufficientResourcesError(
                        "Couldn't find a cpu with full availability!")
                alloc[cpu_key] = 1.
                self.resource_map[cpu_key] = 0.
        else:
            if 'num_cpus' in resource_request:
                if resource_request['num_cpus'] == -1:
                    already_allocated_cpu_keys = alloc.cpus
                    available_cpu_keys = self.cpus
                    remaining_cpu_keys = list(filter(
                        lambda k: k not in already_allocated_cpu_keys,
                        available_cpu_keys))
                    remaining_cpu_keys = list(filter(
                        lambda k: self.resource_map[k] == 1.,
                        remaining_cpu_keys))
                    for cpu_key in remaining_cpu_keys:
                        alloc[cpu_key] = 1.
                        self.resource_map[cpu_key] = 0.
                else:
                    self.release(alloc)
                    raise ValueError(
                        f"'num_cpus' value {resource_request['num_cpus']} "
                        "not supported.")

        return alloc

    def release(self, alloc: ResourceAllocation):
        for key in alloc:
            self.resource_map[key] += alloc[key]


# Create the global resource pool for this process
_resource_pool = ResourcePool()
_context_manager = None


def default_context_loader():
    from dryml.context import ComputeContext
    return ComputeContext


def tf_context_loader():
    from dryml.context.tf import TFComputeContext
    return TFComputeContext


def torch_context_loader():
    from dryml.context.torch import TorchComputeContext
    return TorchComputeContext


context_loaders = {
    'default': default_context_loader,
    'tf': tf_context_loader,
    'torch': torch_context_loader,
}


class context_map(UserDict):
    def __getitem__(self, key):
        try:
            return self.data[key]
        except KeyError:
            ctx_cls = context_loaders[key]()
            self.data[key] = (ctx_cls, make_context_manager(ctx_cls))
            return self.data[key]


contexts = context_map({})


def context():
    global _context_manager
    return _context_manager


def get_context_class(ctx_name):
    global contexts
    return contexts[ctx_name][0]


def get_context_manager(ctx_name):
    global contexts
    return contexts[ctx_name][1]


class ResourcesUnavailableError(Exception):
    """
    Signals a context is unable to allocate necessary resources
    """
    pass


class ContextAlreadyActiveError(Exception):
    """
    Signals a context is already active.
    """
    pass


class WrongContextError(Exception):
    """
    Signals the wrong context is active.
    """
    pass


class NoContextError(Exception):
    """
    Signals there is no context active.
    """
    pass


class ContextIncompatibilityError(Exception):
    """
    Signals there is no single context which satisfies
    all requirements.
    """
    pass


def context_check(ctx_reqs):
    """
    Check if there's a context, and its capabilities
    """
    ctx_mgr = context()
    if ctx_mgr is None:
        raise NoContextError()

    if not ctx_mgr.satisfies(ctx_reqs):
        raise ContextIncompatibilityError(
            f"Context doesn't satisfy requirements {ctx_reqs}")


class ComputeContext(object):
    def __init__(
            self,
            resource_request: Optional[Union[ResourceRequest, dict]] = {}):
        if type(resource_request) is dict:
            resource_request = ResourceRequest(resource_request)
        self.resource_request = resource_request
        self.allocation = None

    def acquire_context(self):
        global _resource_pool
        # Acquire allocation
        self.allocation = _resource_pool.request(self.resource_request)

    def release_context(self):
        global _resource_pool
        # Release allocation
        _resource_pool.release(self.allocation)
        self.allocation = None


class ContextContainer(object):
    """
    Manages the creation of a compute context objects.
    """
    def __init__(self, resource_requests: Optional[dict] = {'default': {}}):
        self.resource_requests = resource_requests
        self.contexts = {}
        self.activated_object_map = {}

    def acquire_context(self):
        # Check that there isn't another manager active already
        global _context_manager
        if _context_manager is not None:
            raise ContextAlreadyActiveError()

        # Acquire needed contexts and resources
        for ctx_name in self.resource_requests:
            # Acquire each context in turn
            ctx_cls = get_context_class(ctx_name)
            ctx = ctx_cls(resource_request=self.resource_requests[ctx_name])
            ctx.acquire_context()
            self.contexts[ctx_name] = ctx

        # Set the global context
        _context_manager = self

    def deactivate_objects(self, save_cache=None):
        # First, we'll build a tree of activated objects.
        from dryml import build_obj_tree
        obj_tree = build_obj_tree(self.activated_object_map.values())

        # Prepare the save caching function
        if save_cache is None:
            from dryml.save_cache import SaveCache
            save_cache = SaveCache()

        # Save objects in the tree depth-first.
        obj_tree.apply_df(
            lambda o: o.save_compute(save_cache=save_cache))
        # Cleanup compute in the tree depth-first.
        obj_tree.apply_df(lambda o: o.compute_cleanup())

        def remove_self(o):
            del self.activated_object_map[id(o)]
        # Remove all objects in the tree from active tracking
        # (Should not fail!)
        obj_tree.apply_df(remove_self)

    def release_context(self):
        # Deactivate each tracked object
        self.deactivate_objects()

        # Release each contained context
        for ctx_name in self.contexts:
            ctx = self.contexts[ctx_name]
            ctx.release_context()

        global _context_manager
        # Remove current_context
        _context_manager = None

    def add_activated_object(self, obj):
        from dryml.object import Object
        if type(obj) is not Object:
            TypeError("Can only activate Objects for computation.")
        self.activated_object_map[id(obj)] = obj

    def remove_activated_object(self, obj):
        from dryml.object import Object
        if type(obj) is not Object:
            TypeError("Can only activate Objects for computation.")
        del self.activated_object_map[id(obj)]

    def contains_activated_object(self, obj):
        from dryml.object import Object
        if type(obj) is not Object:
            TypeError("Can only activate Objects for computation.")
        if id(obj) in self.activated_object_map:
            return True
        else:
            return False

    def satisfies(self, ctx_reqs):
        for ctx_name in ctx_reqs:
            ctx_cls = get_context_class(ctx_name)
            found_satisfier = False
            for c_name in self.contexts:
                c_cls = get_context_class(c_name)
                if ctx_cls in c_cls.mro():
                    if self.contexts[c_name] \
                           .allocation.satisfies(ctx_reqs[ctx_name]):
                        found_satisfier = True
                        break
            if not found_satisfier:
                return False
        return True

    def get_num_gpus_tf(self):
        if 'tf' not in self.contexts:
            raise WrongContextError("No TF Context active.")
        return self.contexts['tf'].allocation.num_gpus

    def get_torch_devices(self):
        if 'torch' not in self.contexts:
            raise WrongContextError("No Torch Context active.")
        return self.contexts['torch'].compute_devices()


def set_context(ctx_reqs):
    """
    Set a context globally. Good for local compute sessions
    """
    ctx_manager = ContextContainer(resource_requests=ctx_reqs)
    ctx_manager.acquire_context()


def make_context_manager(ctx_cls: Type):
    @contextmanager
    def context_manager(*args, **kwargs):
        ctx_obj = ctx_cls(*args, **kwargs)
        ctx_obj.acquire_context()
        try:
            yield ctx_obj
        finally:
            ctx_obj.release_context()

    return context_manager


ContextManager = make_context_manager(ContextContainer)


def consolidate_contexts(ctx_name_list):
    """
    Find a collection of contexts satisfies all listed context
    requirements.
    """

    # Add default context in case no context names are given
    ctx_name_list.append('default')

    # Build list of unique context names
    ctx_name_list = list(set(ctx_name_list))
    ctx_cls_list = list(map(
        lambda name: get_context_class(name),
        ctx_name_list))

    # Get MRO for each class
    ctx_cls_mros = list(map(
        lambda cls: cls.mro(),
        ctx_cls_list))

    # Compute how many context classes from the
    # ctx_cls_list are contained in each mro

    def mro_count(mro, ctx_cls_list):
        return len(list(filter(lambda cls: cls in mro, ctx_cls_list)))

    ctx_cls_mro_contain_count = list(map(
        lambda mro: mro_count(mro, ctx_cls_list),
        ctx_cls_mros))

    # Find context with maximum
    max_v = ctx_cls_mro_contain_count[0]
    max_i = 0
    for i in range(len(ctx_cls_mro_contain_count)):
        if ctx_cls_mro_contain_count[i] > max_v:
            max_i = i
            max_v = ctx_cls_mro_contain_count[i]
    max_ctx_name = ctx_name_list[max_i]

    # form list of leftover contexts which aren't
    # covered by the max context
    max_ctx_mro = ctx_cls_mros[max_i]
    leftover_ctx_names = []

    for i in range(len(ctx_name_list)):
        if ctx_cls_list[i] not in max_ctx_mro:
            leftover_ctx_names.append(ctx_name_list[i])

    if len(leftover_ctx_names) > 0:
        return [max_ctx_name] + consolidate_contexts(leftover_ctx_names)
    else:
        return [max_ctx_name]


def get_context_requirements(objs):
    """
    Set a context appropriate for the object or set of objects
    """
    from dryml import Object

    if issubclass(type(objs), Object):
        objs = [objs]

    if not is_nonstring_iterable(objs):
        raise ValueError(
            "set_appropriate_context only supports single "
            "Objects or an iterable of Objects.")

    ctx_reqs = {}

    for obj in objs:
        obj_reqs = obj.dry_context_requirements()
        for ctx_name in obj_reqs:
            if ctx_name in ctx_reqs:
                ctx_reqs[ctx_name].append(obj_reqs[ctx_name])
            else:
                ctx_reqs[ctx_name] = [obj_reqs[ctx_name]]

    for ctx_name in ctx_reqs:
        ctx_reqs[ctx_name] = combine_requests(ctx_reqs[ctx_name])

    return ctx_reqs



import tensorflow as tf
from typing import Optional, Union
from dryml.context import ComputeContext
from dryml.context.context_tracker import ResourceRequest


class TFComputeContext(ComputeContext):
    def __init__(
            self,
            resource_request: Optional[Union[ResourceRequest, dict]] = {}):
        super().__init__(resource_request=resource_request)
        self.strategy = None
        self.strategy_scope = None

    def acquire_context(self):
        # Let parent handle allocation
        super().acquire_context()

        # Get allocated gpus
        alloc_gpus = self.allocation.gpus

        # check if we need to set memory limits on any of these gpus
        need_mem_limit = False
        for gpu_key in alloc_gpus:
            if self.allocation[gpu_key] < 1.:
                need_mem_limit = True
                break

        # If needed, get actual GPU information
        gpu_info = None
        if need_mem_limit:
            try:
                import GPUtil
                gpu_info = GPUtil.getGPUs()
            except ImportError as e:
                print("Need GPUtil available to set memory limits on gpus.")
                raise e

        def get_gpu_id(key):
            return int(key.split('/')[1])

        # Get ids of gpus assigned. We first need to remove other
        # gpus from being visible.
        visible_gpu_ids = list(map(get_gpu_id, alloc_gpus))
        tf_gpu_objs = tf.config.list_physical_devices(device_type='GPU')
        tf_gpu_objs_restricted = []
        for i in range(len(visible_gpu_ids)):
            tf_gpu_objs_restricted.append(tf_gpu_objs[visible_gpu_ids[i]])
        tf.config.set_visible_devices(tf_gpu_objs_restricted, 'GPU')

        if need_mem_limit:
            # We need to restrict memory for some gpus.
            restrict_gpu_keys = list(filter(
                lambda k: self.allocation[k] < 1.,
                alloc_gpus))

            # Create config objects
            for i in range(len(restrict_gpu_keys)):
                gpu_key = restrict_gpu_keys[i]
                gpu_id = get_gpu_id(gpu_key)
                target_mem = int(
                    gpu_info[gpu_id].memoryTotal*self.allocation[gpu_key])
                tf.config.set_logical_device_configuration(
                    tf_gpu_objs[gpu_id],
                    [tf.config.LogicalDeviceConfiguration(
                        memory_limit=target_mem)])

        logical_devices = tf.config.list_logical_devices('GPU')
        if len(logical_devices) > 1:
            # Get visible logical device names
            gpu_names = list(map(lambda l: l.name, logical_devices))

            # Create
            self.strategy = tf.distribute.MirroredStrategy(gpu_names)
            # Fix improper tear-down on program exit:
            # https://github.com/tensorflow/tensorflow/issues/50487
            import atexit
            atexit.register(
                self.strategy._extended._collective_ops._pool.close)

        if self.strategy is not None:
            self.strategy_scope = self.strategy.scope()
            self.strategy_scope.__enter__()

    def release_context(self):
        if self.strategy_scope is not None:
            self.strategy_scope.__exit__(None, None, None)
            self.strategy_scope = None
            self.strategy = None

        # Let parent handle releasing allocation
        super().release_context()



from dryml.object import Object
from dryml.context import cls_method_compute


@cls_method_compute('__call__')
class Component(Object):
    """
    A Type for an ML component
    """

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()


class TrainFunction(Component):
    def __init__(self):
        self.train_args = ()
        self.train_kwargs = {}

    def __call__(
            self, trainable, train_data,
            train_spec=None, train_callbacks=[]):
        raise NotImplementedError("method must be implemented in a subclass")



from dryml.data.tf.transforms import FuncMap, FuncXMap, FuncYMap, \
    BestCat
from dryml.data.tf.dataset import TFDataset

__all__ = [
    FuncMap,
    FuncXMap,
    FuncYMap,
    TFDataset,
    BestCat,
]



import tensorflow as tf


class keras_train_spec_updater(tf.keras.callbacks.Callback):
    def __init__(self, train_spec, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.train_spec = train_spec

    def on_epoch_end(self, epoch, logs=None):
        # Advance the train spec at end of an epoch
        self.train_spec.advance()


class keras_callback_wrapper(tf.keras.callbacks.Callback):
    def __init__(self, callback, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.callback = callback

    def on_epoch_end(self, epoch, logs=None):
        # Call the callback at the end of the epoch
        self.callback()



from dryml.data.dataset import Dataset
from dryml.models.trainable import Trainable
from dryml.collections import Tuple


class Pipe(Trainable, Tuple):
    """A Sequential processing pipeline modelled after sklearn pipe"""
    def __init__(self, *args, **kwargs):
        # May need to test that elements of the pipes are trainable too
        pass

    def train(
            self, data, *args, train_spec=None,
            train_callbacks=[], **kwargs):
        # get start epoch (only way to do this is to pass a train_spec.
        # Non-zero start epoch can only happen with no train_spec?
        start_epoch = 0
        if train_spec is not None:
            start_epoch = train_spec.level_step()

        last_val = data
        for i in range(len(self)):
            # Get pipe step
            step = self[i]

            if start_epoch > i or step.train_state == Trainable.trained:
                # This part of the training has already been completed.
                pass
            else:
                # We need to train this step.

                # Manage train spec
                if train_spec is not None:
                    train_spec.descend()

                # Perform training
                step.train(
                    last_val, *args, train_spec=train_spec,
                    train_callbacks=train_callbacks, **kwargs)

                # manage train spec
                if train_spec is not None:
                    train_spec.elevate()

            # Eval data for next step in training
            if i < len(self)-1:
                last_val = step.eval(last_val)

            if train_spec is not None and i >= start_epoch:
                train_spec.advance()

        # Call Trainable Train
        super().train(data, *args, **kwargs)

    def eval(self, X: Dataset, *args, **kwargs):
        last_val = X
        for step in self:
            last_val = step.eval(last_val, *args, **kwargs)

        return last_val

    def prep_train(self):
        for step in self:
            step.prep_train()

    def prep_eval(self):
        for step in self:
            step.prep_eval()



import pytest
import dryml
import io
import os
import sys
import time
import importlib

test_objs_text = """import dryml


class SimpleObject(dryml.Object):
    def __init__(self, i, **kwargs):
        self.i = i

    def version(self):
        return {version}

    def __eq__(self, rhs):
        return self.i == rhs.i
"""


def test_object_collect_kwargs_1():
    import objects
    obj = objects.TestClassD3(1, arg1='a', arg2='b')
    assert 1 in obj.dry_args
    assert 'arg1' in obj.dry_kwargs
    assert 'arg2' in obj.dry_kwargs
    assert 'dry_id' not in obj.mdl_kwargs

    obj_def = obj.definition()
    assert 1 in obj_def.args
    assert 'arg1' in obj_def.kwargs
    assert 'arg2' in obj_def.kwargs
    assert 'dry_id' in obj_def.kwargs

    obj2 = obj_def.build()
    assert 1 in obj2.dry_args
    assert 'arg1' in obj2.dry_kwargs
    assert 'arg2' in obj2.dry_kwargs
    assert 'dry_id' not in obj2.mdl_kwargs


def test_save_object_1():
    """
    Test Saving objects through an io buffer
    """
    with open('./tests/test_objs.py', 'w') as f:
        f.write(test_objs_text.format(version=1))

    import test_objs
    importlib.reload(test_objs)

    # Define simple class
    temp_buffer = io.BytesIO()
    obj = test_objs.SimpleObject(10)

    # Test that save to buffer works
    assert obj.save_self(temp_buffer)

    obj2 = dryml.load_object(temp_buffer)

    # Test that restore from buffer creates identical object in this context.
    assert obj == obj2

    assert obj.version() == 1
    assert obj2.version() == 1


@pytest.mark.usefixtures("create_name")
def test_save_object_2(create_name):
    """
    Test Saving objects to a file which doesn't yet exist
    """
    with open('./tests/test_objs.py', 'w') as f:
        f.write(test_objs_text.format(version=1))

    import test_objs
    importlib.reload(test_objs)

    obj = test_objs.SimpleObject(10)

    assert obj.save_self(create_name)

    obj2 = dryml.load_object(create_name)

    assert obj == obj2

    assert obj.version() == 1
    assert obj2.version() == 1


@pytest.mark.usefixtures("create_temp_named_file")
def test_save_object_3(create_temp_named_file):
    """
    Test Saving objects to a file using file which was already created
    """
    with open('./tests/test_objs.py', 'w') as f:
        f.write(test_objs_text.format(version=1))

    import test_objs
    importlib.reload(test_objs)

    obj = test_objs.SimpleObject(10)

    assert obj.save_self(create_temp_named_file)

    obj2 = dryml.load_object(create_temp_named_file)

    assert obj == obj2

    assert obj.version() == 1
    assert obj2.version() == 1


@pytest.mark.usefixtures("create_temp_file")
def test_save_object_4(create_temp_file):
    """
    Test Saving objects to a file using bytes-like file object
    """
    with open('./tests/test_objs.py', 'w') as f:
        f.write(test_objs_text.format(version=1))

    import test_objs
    importlib.reload(test_objs)

    obj = test_objs.SimpleObject(10)

    assert obj.save_self(create_temp_file)

    create_temp_file.flush()
    create_temp_file.seek(0)
    obj2 = dryml.load_object(create_temp_file)

    assert obj == obj2

    assert obj.version() == 1
    assert obj2.version() == 1


@pytest.mark.xfail
@pytest.mark.usefixtures("create_temp_file")
def test_save_object_5(create_temp_file):
    """
    Test Saving objects to a file, then loading in an environment
    without class definition
    """
    # This is currently not possible, or annoyingly difficult:
    # https://github.com/uqfoundation/dill/issues/128
    # Write test objects module, and load it.
    with open('./tests/test_objs.py', 'w') as f:
        f.write(test_objs_text.format(version=1))

    import test_objs
    importlib.reload(test_objs)

    # Create object and save
    obj = test_objs.SimpleObject(10)

    assert obj.save_self(create_temp_file)

    # Delete test_objs source and module from sys
    if os.path.exists('./tests/test_objs.py'):
        os.remove('./tests/test_objs.py')

    del test_objs
    if 'test_objs' in sys.modules:
        del sys.modules['test_objs']

    # Rewind file
    create_temp_file.flush()
    create_temp_file.seek(0)

    obj2 = dryml.load_object(create_temp_file)

    assert obj == obj2

    assert obj.version() == 1
    assert obj2.version() == 1


@pytest.mark.usefixtures("create_temp_named_file")
def test_save_object_6(create_temp_named_file):
    """
    Test object default metadata saving
    """
    with open('./tests/test_objs.py', 'w') as f:
        f.write(test_objs_text.format(version=1))

    import test_objs
    importlib.reload(test_objs)

    desc_str = 'Test Description'
    obj = test_objs.SimpleObject(10, dry_metadata={'description': desc_str})
    orig_creation_time = obj.dry_metadata['creation_time']

    assert obj.save_self(create_temp_named_file)

    obj2 = dryml.load_object(create_temp_named_file)

    assert obj == obj2

    assert obj.version() == 1
    assert obj2.version() == 1

    assert orig_creation_time == obj2.dry_metadata['creation_time']
    assert desc_str == obj2.dry_metadata['description']


def test_basic_object_def_update_1():
    def build_and_save_obj_1():
        time.sleep(1.1)
        with open('tests/test_objs.py', 'w') as f:
            f.write(test_objs_text.format(version=1))

        import test_objs
        importlib.reload(test_objs)

        obj = test_objs.SimpleObject(10)

        buffer = io.BytesIO()

        assert obj.save_self(buffer)

        return obj, buffer

    obj1, buffer = build_and_save_obj_1()

    def build_obj_2(buffer):
        time.sleep(1.1)
        with open('tests/test_objs.py', 'w') as f:
            f.write(test_objs_text.format(version=2))
        # Sleep to invalidate the cache.

        obj2 = dryml.load_object(buffer, update=True, reload=True)

        return obj2

    obj2 = build_obj_2(buffer)

    assert obj1 == obj2

    assert obj1.version() == 1
    assert obj2.version() == 2


@pytest.mark.usefixtures("create_name")
def test_basic_object_def_update_2(create_name):
    def build_and_save_obj_1():
        time.sleep(1.1)
        with open('tests/test_objs.py', 'w') as f:
            f.write(test_objs_text.format(version=1))
        import test_objs
        importlib.reload(test_objs)

        obj = test_objs.SimpleObject(10)

        assert obj.save_self(create_name)

        return obj

    obj1 = build_and_save_obj_1()

    def build_obj_2():
        # Sleep to invalidate the cache.
        time.sleep(1.1)
        with open('tests/test_objs.py', 'w') as f:
            f.write(test_objs_text.format(version=2))

        obj2 = dryml.load_object(create_name, update=True, reload=True)

        return obj2

    obj2 = build_obj_2()

    assert obj1 == obj2

    assert obj1.version() == 1
    assert obj2.version() == 2


def test_object_build_from_def_1():
    """
    Test that an object definition with no id results in an object with an id.
    """

    import objects
    obj = dryml.ObjectDef(
        objects.TestClassB,
        1,
        base_msg='Test').build()

    assert 'dry_id' in obj.dry_kwargs


def test_object_args_passing_1():
    import objects as objs

    obj = objs.TestClassB(1, base_msg="Test1")

    assert obj.dry_args == (1,)


@pytest.mark.usefixtures("create_name")
def test_object_args_passing_2(create_name):
    import objects as objs

    obj = objs.TestClassB(1, base_msg="Test1")

    dryml.save_object(obj, create_name)

    obj_loaded = dryml.load_object(create_name)

    assert obj_loaded.dry_args == (1,)


def test_object_args_passing_3():
    """
    Test passing nested dryobjects as arguments
    """
    import objects as objs

    obj1 = objs.TestNest(10)

    obj2 = objs.TestNest(obj1)

    obj3 = objs.TestNest(obj2)

    obj1_def = obj1.definition()

    obj1_cpy = obj1_def.build()

    assert obj1.definition() == obj1_cpy.definition()
    assert obj1 is not obj1_cpy

    obj2_def = obj2.definition()

    obj2_cpy = obj2_def.build()

    assert obj2.definition() == obj2_cpy.definition()
    assert obj2 is not obj2_cpy
    assert obj2.A is not obj2_cpy.A
    assert type(obj2.A) is type(obj1)
    assert type(obj2_cpy.A) is type(obj1)

    obj3_cpy = obj3.definition().build()

    assert obj3.definition() == obj3_cpy.definition()
    assert obj3 is not obj3_cpy
    assert obj3.A is not obj3_cpy.A
    assert type(obj3.A) is type(obj2)
    assert type(obj3_cpy.A) is type(obj2)
    assert obj3.A.A is not obj3_cpy.A.A
    assert type(obj3.A.A) is type(obj1)
    assert type(obj3_cpy.A.A) is type(obj1)


def test_object_args_passing_4():
    """
    Test passing nested dryobjects as arguments, within a list
    """
    import objects as objs

    obj1 = objs.TestNest(10)

    obj2 = objs.TestNest([obj1])

    obj1_cpy = obj1.definition().build()

    assert obj1.definition() == obj1_cpy.definition()
    assert obj1 is not obj1_cpy

    obj2_cpy = obj2.definition().build()

    assert obj2.definition() == obj2_cpy.definition()
    assert obj2 is not obj2_cpy
    assert obj2.A is not obj2_cpy.A
    assert type(obj2.A[0]) is type(obj1)
    assert type(obj2_cpy.A[0]) is type(obj1)


def test_object_args_passing_5():
    """
    Test passing nested dryobjects as arguments, within a nested list
    """
    import objects as objs

    obj1 = objs.TestNest(10)

    obj2 = objs.TestNest([[obj1]])

    obj1_cpy = obj1.definition().build()

    assert obj1.definition() == obj1_cpy.definition()
    assert obj1 is not obj1_cpy

    obj2_cpy = obj2.definition().build()

    assert obj2.definition() == obj2_cpy.definition()
    assert obj2 is not obj2_cpy
    assert obj2.A is not obj2_cpy.A
    assert type(obj2.A[0][0]) is type(obj1)
    assert type(obj2_cpy.A[0][0]) is type(obj1)


def test_object_args_passing_6():
    """
    Test passing nested dryobjects as arguments, within a dict
    """
    import objects as objs

    obj1 = objs.TestNest(10)

    obj2 = objs.TestNest({'A': obj1})

    obj1_cpy = obj1.definition().build()

    assert obj1.definition() == obj1_cpy.definition()
    assert obj1 is not obj1_cpy

    obj2_cpy = obj2.definition().build()

    assert obj2.definition() == obj2_cpy.definition()
    assert obj2 is not obj2_cpy
    assert obj2.A is not obj2_cpy.A
    assert type(obj2.A['A']) is type(obj1)
    assert type(obj2_cpy.A['A']) is type(obj1)


def test_object_args_passing_7():
    """
    Test passing nested dryobjects as arguments, within a dict with a list
    """
    import objects as objs

    obj1 = objs.TestNest(10)

    obj2 = objs.TestNest({'A': [[obj1]]})

    obj1_cpy = obj1.definition().build()

    assert obj1.definition() == obj1_cpy.definition()
    assert obj1 is not obj1_cpy

    obj2_cpy = obj2.definition().build()

    assert obj2.definition() == obj2_cpy.definition()
    assert obj2 is not obj2_cpy
    assert obj2.A is not obj2_cpy.A
    assert type(obj2.A['A'][0][0]) is type(obj1)
    assert type(obj2_cpy.A['A'][0][0]) is type(obj1)


def test_object_config_1():
    import objects as objs

    obj = objs.HelloStr(msg="Test")
    msg = obj.get_message()
    assert msg == "Hello! Test"

    obj = objs.HelloInt(msg=10)
    msg = obj.get_message()
    assert msg == "Hello! 10"


def test_object_hash_1():
    "Test that object hashes are unique within classes"
    import objects as objs
    obj1 = objs.HelloStr(msg="Test")
    obj2 = objs.HelloStr(msg="Test")
    assert obj1.definition().get_individual_id() != \
        obj2.definition().get_individual_id()


def test_object_hash_2():
    "Test that object hashes are are same for two elements of the same class"
    import objects as objs
    obj1 = objs.HelloStr(msg="Test")
    obj2 = objs.HelloStr(msg="Test")
    assert obj1.definition().get_category_id() == \
        obj2.definition().get_category_id()


@pytest.mark.usefixtures("create_name")
def test_object_hash_3(create_name):
    "Test that object hashes are the same after saving and restoring"
    import objects as objs
    obj1 = objs.HelloStr(msg="Test")
    assert obj1.save_self(create_name)

    obj2 = dryml.load_object(create_name)
    assert obj1.definition().get_category_id() == \
        obj2.definition().get_category_id()


@pytest.mark.usefixtures("create_name")
def test_object_hash_4(create_name):
    "Test that loaded objects are identical hash wise"
    import objects as objs
    obj1 = objs.HelloStr(msg="Test")
    assert obj1.save_self(create_name)

    obj2 = dryml.load_object(create_name)
    assert obj1.definition().get_individual_id() == \
        obj2.definition().get_individual_id()


@pytest.mark.usefixtures("create_name")
def test_object_file_hash_1(create_name):
    "Test that object hashes are the same after saving and restoring"
    import objects as objs
    obj1 = objs.HelloStr(msg="Test")
    assert obj1.save_self(create_name)

    with dryml.ObjectFile(create_name) as dry_file:
        assert obj1.definition().get_category_id() == \
            dry_file.definition().get_category_id()


@pytest.mark.usefixtures("create_name")
def test_object_file_hash_2(create_name):
    "Test that loaded objects are identical hash wise"
    import objects as objs
    obj1 = objs.HelloStr(msg="Test")
    assert obj1.save_self(create_name)

    with dryml.ObjectFile(create_name) as dry_file:
        assert obj1.definition().get_individual_id() == \
            dry_file.definition().get_individual_id()


@pytest.mark.usefixtures("create_name")
def test_object_file_hash_3(create_name):
    "Test that object and object factory hashes are the same"
    import objects as objs
    obj1 = objs.HelloStr(msg="Test")
    assert obj1.save_self(create_name)

    f = dryml.ObjectFactory(dryml.ObjectDef(
        objs.HelloStr, msg="Test"))

    assert obj1.definition().get_category_id() == \
        f.obj_def.get_category_id()


def test_change_obj_cls_1():
    "Test that we can change an object's class"
    import objects as objs
    obj1 = objs.TestClassA(item=[5])
    obj2 = dryml.change_object_cls(obj1, objs.TestClassA2)

    assert type(obj2) is objs.TestClassA2
    assert obj1.dry_kwargs['item'] == obj2.dry_kwargs['item']


def test_object_def_1():
    import objects
    obj_def = dryml.ObjectDef(objects.HelloInt, msg=10)
    other_def = dryml.ObjectDef(
        objects.HelloInt,
        msg=10)

    assert obj_def['cls'] is other_def['cls']
    assert obj_def['dry_args'] == obj_def['dry_args']
    assert obj_def['dry_kwargs'] == obj_def['dry_kwargs']


def test_object_def_2():
    import objects
    obj_def = dryml.ObjectDef(objects.HelloInt, msg=10)

    new_obj = obj_def.build()

    assert isinstance(new_obj, objects.HelloInt)
    assert new_obj.dry_kwargs['msg'] == 10


def test_object_fac_1():
    import objects

    obj_fac = dryml.ObjectFactory(
        dryml.ObjectDef(objects.HelloInt, msg=10))

    obj = obj_fac()

    assert isinstance(obj, objects.HelloInt)
    assert obj.dry_kwargs['msg'] == 10


@pytest.mark.usefixtures("create_temp_named_file")
def test_object_save_restore_1(create_temp_named_file):
    """
    We test save and restore of nested objects through arguments
    """
    import objects

    # Create the data containing objects
    data_obj1 = objects.TestClassC2(10)
    data_obj1.set_val(20)

    data_obj2 = objects.TestClassC2(20)
    data_obj2.set_val(40)

    # Enclose them in another object
    obj = objects.TestClassC(data_obj1, B=data_obj2)

    assert obj.save_self(create_temp_named_file)

    obj2 = dryml.load_object(create_temp_named_file)

    assert obj.definition() == obj2.definition()
    assert obj.A.data == obj2.A.data
    assert obj.B.data == obj2.B.data


@pytest.mark.usefixtures("create_temp_named_file")
def test_object_save_restore_2(create_temp_named_file):
    """
    We test save and restore of nested objects through arguments
    This time, we make sure identical objects are loaded as
    the same object.
    """
    import objects

    # Create the data containing objects
    data_obj1 = objects.TestClassC2(10)
    data_obj1.set_val(20)

    # Enclose them in another object
    obj = objects.TestClassC(data_obj1, B=data_obj1)

    assert obj.save_self(create_temp_named_file)

    # Load the object from the file
    obj2 = dryml.load_object(create_temp_named_file)

    assert obj.definition() == obj2.definition()
    assert obj.A is obj.B


@pytest.mark.usefixtures("create_temp_named_file")
def test_object_save_restore_3(create_temp_named_file):
    """
    We test save and restore of nested objects through arguments
    Deeper nesting
    """
    import objects

    # Create the data containing objects
    data_obj1 = objects.TestClassC2(10)
    data_obj1.set_val(20)

    data_obj2 = objects.TestClassC2(20)
    data_obj2.set_val(40)

    data_obj3 = objects.TestClassC2('test')
    data_obj3.set_val('test')

    data_obj4 = objects.TestClassC2(0.5)
    data_obj4.set_val(30.5)

    obj1 = objects.TestClassC(data_obj1, B=data_obj2)
    obj2 = objects.TestClassC(data_obj3, B=data_obj4)

    # Enclose them in another object
    obj = objects.TestClassC(obj1, B=obj2)

    assert obj.save_self(create_temp_named_file)

    # Load the object from the file
    obj2 = dryml.load_object(create_temp_named_file)

    assert obj.definition() == obj2.definition()
    assert obj.A.A.data == obj2.A.A.data
    assert obj.A.B.data == obj2.A.B.data
    assert obj.B.A.data == obj2.B.A.data
    assert obj.B.B.data == obj2.B.B.data


def test_object_save_restore_4():
    """
    Test saving/restoring arguments/kwargs
    """

    import objects

    # Create the data containing objects
    data_obj1 = objects.TestClassC2(10)
    data_obj1.set_val(20)

    data_obj2 = objects.TestClassC2(20)
    data_obj2.set_val(40)

    data_obj3 = objects.TestClassC2('test')
    data_obj3.set_val('test')

    data_obj4 = objects.TestClassC2(0.5)
    data_obj4.set_val(30.5)

    obj1 = objects.TestClassC(data_obj1, B=data_obj2)
    obj2 = objects.TestClassC(data_obj3, B=data_obj4)

    from dryml.object import DryObjectPlaceholder, \
        prep_args_kwargs, reconstruct_args_kwargs

    args = (obj1, obj2)

    (args, kwargs), ph = prep_args_kwargs(args, {})

    assert type(args[0]) is DryObjectPlaceholder
    assert type(args[1]) is DryObjectPlaceholder

    reconstruct_args_kwargs(args, kwargs, ph)

    assert type(args[0]) is objects.TestClassC
    assert type(args[1]) is objects.TestClassC

    assert obj1.A.data == args[0].A.data
    assert obj1.B.data == args[0].B.data
    assert obj2.A.data == args[1].A.data
    assert obj2.B.data == args[1].B.data


def test_object_save_restore_5():
    """
    Test saving/restoring arguments/kwargs
    """

    import objects

    # Create the data containing objects
    model_obj = objects.TestNest(10)
    opt_obj = objects.TestNest3(20, model=model_obj)
    loss_obj = objects.TestNest2(A='func')
    train_fn_obj = objects.TestNest3(
        optimizer=opt_obj,
        loss=loss_obj,
        epochs=10)

    trainable_obj = objects.TestNest3(
        model=model_obj,
        train_fn=train_fn_obj
    )

    from dryml.object import DryObjectPlaceholder, \
        prep_args_kwargs, reconstruct_args_kwargs

    args = (trainable_obj,)

    (args, kwargs), ph = prep_args_kwargs(args, {})

    assert type(args[0]) is DryObjectPlaceholder

    reconstruct_args_kwargs(args, kwargs, ph)

    recon_trainable_obj = args[0]
    assert type(recon_trainable_obj) is objects.TestNest3

    assert recon_trainable_obj['model'] is \
        recon_trainable_obj['train_fn']['optimizer']['model']
    assert recon_trainable_obj['train_fn']['epochs'] == 10
    assert recon_trainable_obj['model'].A == 10
    assert recon_trainable_obj['train_fn']['optimizer'][0] == 20


def test_nested_def_build_1():
    """
    Test nested definitions build appropriately.
    """

    import objects

    data_def1 = dryml.ObjectDef(objects.TestNest2, A=1)
    data_def2 = dryml.ObjectDef(objects.TestNest2, A=2)

    data_def = dryml.ObjectDef(objects.TestClassC, data_def1, B=data_def1)
    obj = data_def.build()
    assert obj.A.A == 1
    assert obj.B.A == 1
    assert obj.A is obj.B

    data_def = dryml.ObjectDef(objects.TestClassC, data_def2, B=data_def2)
    obj = data_def.build()
    assert obj.A.A == 2
    assert obj.B.A == 2
    assert obj.A is obj.B

    data_def = dryml.ObjectDef(objects.TestClassC, data_def1, B=data_def2)
    obj = data_def.build()
    assert obj.A.A == 1
    assert obj.B.A == 2
    assert obj.A is not obj.B


def test_build_crash_1():
    import objects

    # First wrong definition
    test_def_1 = dryml.ObjectDef(objects.TestClassG1, 1, tmp='5')

    try:
        test_def_1.build()
    except TypeError:
        # This definition will throw a TypeError.
        pass

    print(dryml.config.build_cache)

    # Second corrected definition
    test_def_2 = dryml.ObjectDef(objects.TestClassG1, 1)

    obj1 = test_def_2.build()
    obj2 = test_def_2.build()

    assert obj1.dry_id != obj2.dry_id



from dryml.data import Dataset, \
    NumpyDataset, util
from typing import Callable
import torch
import numpy as np
from dryml.data.util import taker, skiper, nested_batcher


class TorchIterableDatasetWrapper(torch.utils.data.IterableDataset):
    def __init__(
            self, iterable_gen: Callable):
        self.iterable_gen = iterable_gen

    def __iter__(self):
        return iter(self.iterable_gen())


class TorchDataset(Dataset):
    def __init__(
            self, in_ds: torch.utils.data.Dataset, indexed=False,
            supervised=False, batch_size=None, size=None):

        super().__init__(
            indexed=indexed, supervised=supervised,
            batch_size=batch_size)
        self.ds = in_ds
        if size is None:
            size = np.nan
        self.size = size

    def as_indexed(self, start=0) -> Dataset:
        """
        If not already indexed, return a version of this dataset
        which is indexed.
        """
        if self.indexed:
            return self
        else:
            if not self.batched:
                def enumerate_dataset(gen_func, start=0):
                    i = start
                    it = iter(gen_func())
                    while True:
                        try:
                            d = next(it)
                        except StopIteration:
                            return
                        yield (torch.tensor(i), d)
                        i += 1

                torch_ds = TorchIterableDatasetWrapper(
                    lambda: enumerate_dataset(self.data_gen, start=start))

                return TorchDataset(
                    torch_ds,
                    indexed=True,
                    supervised=self.supervised,
                    batch_size=self.batch_size,
                    size=self.size)
            else:
                def enumerate_dataset(gen_func, start=0):
                    it = iter(gen_func())
                    i = start
                    while True:
                        try:
                            d = next(it)
                        except StopIteration:
                            return
                        from dryml.data.util import get_data_batch_size
                        batch_size = get_data_batch_size(d)
                        idx = torch.tensor(list(range(i, i+batch_size)))
                        i += batch_size
                        yield (idx, d)

                torch_ds = TorchIterableDatasetWrapper(
                    lambda: enumerate_dataset(
                        lambda: iter(self.ds),
                        start=start))

                return TorchDataset(
                    torch_ds,
                    indexed=True,
                    supervised=self.supervised,
                    batch_size=self.batch_size,
                    size=self.size)

    @property
    def data_gen(self):
        return lambda: self.ds

    def data(self):
        """
        Get the internal dataset
        """
        return self.ds

    def batch(self, batch_size=32, drop_remainder=True) -> Dataset:
        """
        Batch this data
        """
        if self.batched:
            if self.batch_size != batch_size:
                return self.unbatch().batch(batch_size=batch_size)
            else:
                return self
        else:

            obj = TorchIterableDatasetWrapper(
                lambda: nested_batcher(
                    self.data_gen, batch_size,
                    lambda e: torch.stack(e, dim=0),
                    drop_remainder=drop_remainder))

            return TorchDataset(
                obj,
                indexed=self.indexed,
                supervised=self.supervised,
                batch_size=batch_size,
                size=self.size)

    def unbatch(self) -> Dataset:
        """
        Unbatch this data
        """
        if not self.batched:
            return self
        else:
            obj = TorchIterableDatasetWrapper(
                lambda: util.nested_unbatcher(self.data_gen))
            return TorchDataset(
                obj,
                indexed=self.indexed,
                supervised=self.supervised,
                size=self.size)

    def map(self, func: Callable = None):
        obj = TorchIterableDatasetWrapper(
            lambda: map(func, self.data_gen()))

        return TorchDataset(
            obj,
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size,
            size=self.size)

    def take(self, n):
        """
        Take only a specific number of examples
        """

        new_size = self.size
        if new_size is np.nan:
            new_size = n
        elif new_size is np.inf:
            new_size = n
        else:
            if new_size > n:
                new_size = n

        obj = TorchIterableDatasetWrapper(
            lambda: taker(self.data_gen, n))

        return TorchDataset(
            obj,
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size,
            size=new_size)

    def skip(self, n):
        """
        Skip a specific number of examples
        """

        new_size = self.size
        if new_size is not np.nan and new_size is not np.inf:
            if n > new_size:
                new_size = 0
            else:
                new_size -= n

        obj = TorchIterableDatasetWrapper(
            lambda: skiper(self.data_gen, n))

        return TorchDataset(
            obj,
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=self.batch_size,
            size=new_size)

    def __len__(self):
        """
        Get length of dataset. Will return Infinite if infinite,
        and unknown if it can't be determined.
        """
        return self.size

    def numpy(self):
        dataset = self.map_el(lambda el: el.detach().numpy())
        return NumpyDataset(
            dataset.data_gen,
            indexed=dataset.indexed,
            supervised=dataset.supervised,
            batch_size=dataset.batch_size,
            size=dataset.size)

    def tf(self):
        return self.numpy().tf()

    def torch(self):
        return self

    def shuffle(self, buffer_size, seed=None):
        # create generator of data
        def shuffler():
            # Create new generator
            rng = np.random.default_rng(seed=seed)

            # create iterator on unbatched data
            ds_iter = iter(self.unbatch())

            # create buffer and fill it
            el_buffer = []

            # Fill the buffer
            for idx in range(buffer_size):
                try:
                    el_buffer.append(next(ds_iter))
                except StopIteration:
                    break

            while True:
                # Compute size of buffer, and break out if nothing left
                num_in_buffer = len(el_buffer)
                if num_in_buffer == 0:
                    break

                if num_in_buffer > 1:
                    idx = rng.integers(low=0, high=num_in_buffer-1)
                else:
                    idx = 0

                yield el_buffer.pop(idx)

                # Try to refill the buffer here
                try:
                    el_buffer.append(next(ds_iter))
                except StopIteration:
                    pass

        ds = TorchIterableDatasetWrapper(
            shuffler)

        return TorchDataset(
            ds,
            indexed=self.indexed,
            supervised=self.supervised,
            batch_size=None)



from dryml.data.torch.dataset import TorchDataset, TorchIterableDatasetWrapper
import dryml.data.torch.transforms as transforms

__all__ = [
    TorchDataset,
    TorchIterableDatasetWrapper,
    transforms,
]



import dryml


def test_resource_pool_1():
    pool_obj = dryml.context.ResourcePool(num_cpus=3, num_gpus=3, _test=True)

    req = dryml.context.ResourceRequest(num_gpus=1)
    alloc = pool_obj.request(req)

    assert alloc.satisfies(req)

    pool_obj.release(alloc)

    req2 = dryml.context.ResourceRequest({'gpu/1': 0.5})
    alloc = pool_obj.request(req2)

    alloc2 = pool_obj.request(req2)

    assert alloc.satisfies(req2)
    assert alloc2.satisfies(req2)

    pool_obj.release(alloc)
    pool_obj.release(alloc2)

    for key in pool_obj.resource_map:
        assert pool_obj.resource_map[key] == 1.


def test_context_check_1():
    with dryml.context.ContextManager(
            resource_requests={'default': {'num_cpus': 1}}):
        dryml.context.context_check({'default': {}})


def test_context_check_2():
    with dryml.context.ContextManager(
            resource_requests={'default': {'num_cpus': 1}}):
        try:
            dryml.context.context_check({'tf': {}})
            assert False
        except dryml.context.ContextIncompatibilityError:
            pass



import dryml
import objects
import mc
import numpy as np
import pickle
import pytest


def test_compute_context_consistency_1():
    obj = objects.TestClassE()

    assert obj.__dry_compute_data__ is None

    @dryml.compute
    def func1(obj):
        obj.set_val(20)

    func1(obj)

    assert obj.__dry_compute_data__ is None


def test_compute_context_consistency_2():
    obj = objects.TestClassE()

    assert obj.__dry_compute_data__ is None

    @dryml.compute_context(ctx_update_objs=True)
    def func1(obj):
        obj.set_val(20)

    func1(obj)

    assert obj.__dry_compute_data__ is not None

    @dryml.compute
    def func2(obj):
        assert obj.val == 20


# Model and definition generators


def sklearn_regressor(definition=False):
    sklearn = pytest.importorskip('sklearn')
    import dryml.models.sklearn
    import sklearn.neighbors

    # Create model
    model = dryml.models.sklearn.Trainable(
        model=dryml.models.sklearn.RegressionModel(
            sklearn.neighbors.KNeighborsRegressor,
            n_neighbors=5,
            weights='uniform',
            algorithm='ball_tree'),
        train_fn=dryml.models.sklearn.BasicTraining(
            shuffle=True,
            shuffle_buffer_size=1000,
            ))

    if definition:
        return model.definition().get_cat_def(recursive=True)
    else:
        return model


def tf_regressor(definition=False):
    tf = pytest.importorskip('tensorflow')
    import dryml.models.tf
    import tensorflow as tf

    # Create model
    model = dryml.models.tf.keras.Trainable(
        model=dryml.models.tf.keras.SequentialFunctionalModel(
            input_shape=(1,),
            layer_defs=[
                ('Dense', {'units': 32, 'activation': 'relu'}),
                ('Dense', {'units': 32, 'activation': 'relu'}),
                ('Dense', {'units': 1, 'activation': 'linear'})]),
        optimizer=dryml.models.tf.Wrapper(tf.keras.optimizers.Adam),
        loss=dryml.models.tf.Wrapper(
            tf.keras.losses.MeanSquaredError),
        train_fn=dryml.models.tf.keras.BasicTraining(epochs=1),
        metrics=[
            dryml.models.tf.Wrapper(
                tf.keras.metrics.MeanSquaredError)
        ])

    if definition:
        return model.definition().get_cat_def(recursive=True)
    else:
        return model


def xgb_regressor(definition=False):
    xgboost = pytest.importorskip('xgboost')
    print(xgboost.__version__)
    import dryml.models.sklearn
    import dryml.models.xgb
    import dryml.data.transforms

    model = dryml.models.sklearn.Trainable(
        model=dryml.models.xgb.RegressionModel(),
        train_fn=dryml.models.sklearn.BasicTraining(
            shuffle=True,
            shuffle_buffer_size=1000,),
    )

    flattener = dryml.data.transforms.Flatten()

    model = dryml.models.Pipe(flattener, model)

    if definition:
        return model.definition().get_cat_def(recursive=True)
    else:
        return model


def torch_regressor(definition=False):
    torch = pytest.importorskip('torch')
    import dryml.models.torch
    import dryml.data.torch

    # Create model
    model = dryml.models.torch.generic.Sequential(
        layer_defs=[
            (torch.nn.LazyLinear, (32,), {}),
            (torch.nn.ReLU, (), {}),
            (torch.nn.LazyLinear, (32,), {}),
            (torch.nn.ReLU, (), {}),
            (torch.nn.LazyLinear, (1,), {})]
        )

    model = dryml.models.torch.generic.Trainable(
        model=model,
        train_fn=dryml.models.torch.generic.BasicTraining(
            epochs=1,
            optimizer=dryml.models.torch.generic.TorchOptimizer(
                torch.optim.Adam, model),
            loss=dryml.models.torch.base.Wrapper(
                torch.nn.CrossEntropyLoss)
            ),
        )

    pipe = dryml.models.Pipe(
        dryml.data.transforms.Cast(mode='X', dtype='float32'),
        model,
        dryml.data.torch.transforms.TorchDevice())

    if definition:
        return pipe.definition().get_cat_def(recursive=True)
    else:
        return pipe


regressor_funcs = [
    sklearn_regressor,
    tf_regressor,
    xgb_regressor,
    torch_regressor,
]


def sklearn_classifier(num_dims, num_classes, definition=False):
    sklearn = pytest.importorskip('sklearn')
    import dryml.models.sklearn
    import sklearn.neighbors
    import dryml.data.transforms

    model = dryml.models.sklearn.Trainable(
        model=dryml.models.sklearn.ClassifierModel(
            sklearn.neighbors.KNeighborsClassifier,
            n_neighbors=5,
            algorithm='ball_tree'),
        train_fn=dryml.models.sklearn.BasicTraining(
            shuffle=True,
            shuffle_buffer_size=1000,)
        )

    best_cat = dryml.data.transforms.BestCat()

    model = dryml.models.Pipe(model, best_cat)

    if definition:
        return model.definition().get_cat_def(recursive=True)
    else:
        return model


def tf_classifier(num_dims, num_classes, definition=False):
    tf = pytest.importorskip('tensorflow')
    import dryml.models.tf
    import dryml.data.transforms

    model = dryml.models.tf.keras.Trainable(
        model=dryml.models.tf.keras.SequentialFunctionalModel(
            input_shape=(num_dims,),
            layer_defs=[
                ('Dense', {'units': 32, 'activation': 'relu'}),
                ('Dense', {'units': 32, 'activation': 'relu'}),
                ('Dense', {'units': num_classes, 'activation': 'softmax'}),
            ]),
        optimizer=dryml.models.tf.Wrapper(tf.keras.optimizers.Adam),
        loss=dryml.models.tf.Wrapper(
            tf.keras.losses.SparseCategoricalCrossentropy),
        train_fn=dryml.models.tf.keras.BasicTraining(epochs=1))

    best_cat = dryml.data.transforms.BestCat()

    model = dryml.models.Pipe(model, best_cat)

    if definition:
        model_def = model.definition().get_cat_def(recursive=True)
        return model_def
    else:
        return model


def xgb_classifier(num_dims, num_classes, definition=False):
    xgboost = pytest.importorskip('xgboost')
    print(xgboost.__version__)
    import dryml.models.sklearn
    import dryml.models.xgb
    import dryml.data.transforms

    model = dryml.models.sklearn.Trainable(
        model=dryml.models.xgb.ClassifierModel(),
        train_fn=dryml.models.sklearn.BasicTraining(
            shuffle=True,
            shuffle_buffer_size=1000),
    )

    flattener = dryml.data.transforms.Flatten()
    best_cat = dryml.data.transforms.BestCat()

    model = dryml.models.Pipe(flattener, model, best_cat)

    if definition:
        return model.definition().get_cat_def(recursive=True)
    else:
        return model


def torch_classifier(num_dims, num_classes, definition=False):
    torch = pytest.importorskip('torch')
    import dryml.models.torch
    import dryml.data.torch
    import dryml.data.transforms
    import dryml.data.torch.transforms

    model_obj = dryml.models.torch.generic.Sequential(
        layer_defs=[
            (torch.nn.LazyLinear, (32,), {}),
            (torch.nn.ReLU, (), {}),
            (torch.nn.LazyLinear, (32,), {}),
            (torch.nn.ReLU, (), {}),
            (torch.nn.LazyLinear, (num_classes,), {}),
        ])

    model = dryml.models.torch.generic.Trainable(
        model=model_obj,
        train_fn=dryml.models.torch.generic.BasicTraining(
            epochs=1,
            optimizer=dryml.models.torch.generic.TorchOptimizer(
                torch.optim.Adam,
                model_obj),
            loss=dryml.models.torch.base.Wrapper(
                torch.nn.CrossEntropyLoss),
            )
        )

    best_cat = dryml.data.transforms.BestCat()

    model = dryml.models.Pipe(
        dryml.data.transforms.Cast(mode='X', dtype='float32'),
        model,
        dryml.data.torch.transforms.TorchDevice(),
        best_cat)

    if definition:
        return model.definition().get_cat_def(recursive=True)
    else:
        return model


classifier_funcs = [
    sklearn_classifier,
    tf_classifier,
    xgb_classifier,
    torch_classifier,
]


def gen_dataset_1():
    num_train = 2000
    train_data = mc.gen_dataset_1(num_examples=num_train)
    num_test = 500
    test_data = mc.gen_dataset_1(num_examples=num_test)

    return {
        'train': (num_train, train_data),
        'test': (num_test, test_data),
    }


# Generate dataset 1
dataset_1 = gen_dataset_1()


@pytest.mark.parametrize(
    "model_gen", regressor_funcs)
def test_train_supervised_1_context(model_gen):
    # Generate the model
    model = model_gen(definition=False)

    # Fetch data.
    (num_train, train_data) = dataset_1['train']
    (num_test, test_data) = dataset_1['test']

    def test_model(test_data, model):
        import dryml.metrics

        test_ds = dryml.data.NumpyDataset(test_data, supervised=True)

        return dryml.metrics.mean_squared_error(model, test_ds)

    @dryml.compute_context(
        ctx_use_existing_context=False,
        ctx_update_objs=True)
    def train_method(train_data, test_data, model):
        # Create datasets
        train_ds = dryml.data.NumpyDataset(
            train_data,
            supervised=True)

        # Train model
        model.prep_train()
        model.train(train_ds)

        return test_model(test_data, model)

    @dryml.compute_context(ctx_use_existing_context=False)
    def test_method(test_data, model):
        return test_model(test_data, model)

    train_loss = train_method(train_data, test_data, model)
    test_loss = test_method(test_data, model)

    assert train_loss == test_loss


@pytest.mark.usefixtures("create_temp_dir")
@pytest.mark.parametrize(
    "model_gen", regressor_funcs)
def test_train_supervised_1_context_repo(create_temp_dir, model_gen):
    # Fetch data.
    (num_train, train_data) = dataset_1['train']
    (num_test, test_data) = dataset_1['test']

    model_def = model_gen(definition=True)

    def test_model(test_data, model):
        import dryml.metrics
        test_ds = dryml.data.NumpyDataset(test_data, supervised=True)

        return dryml.metrics.mean_squared_error(model, test_ds)

    model_reqs = model_def.build().dry_context_requirements()

    @dryml.compute_context(
        ctx_context_reqs=model_reqs,
        ctx_use_existing_context=False,
        ctx_update_objs=True)
    def train_method(train_data, test_data, model_def, work_dir):
        # Create repo
        import dryml
        repo = dryml.Repo(directory=work_dir)

        # Create datasets
        train_ds = dryml.data.NumpyDataset(
            train_data,
            batch_size=num_train,
            supervised=True)

        # Build and fetch model
        model = model_def.build(repo=repo)

        # Train model
        model.prep_train()
        model.train(train_ds)

        # Save model to repo.
        repo.save(model)

        return test_model(test_data, model)

    @dryml.compute_context(
        ctx_context_reqs=model_reqs,
        ctx_use_existing_context=False)
    def test_method(test_data, model_def, work_dir):
        import dryml
        # Create repo
        repo = dryml.Repo(directory=work_dir)

        # Fetch model based on definition
        model = dryml.utils.head(repo.get(model_def, build_missing_def=False))

        return test_model(test_data, model)

    train_loss = train_method(
        train_data, test_data, model_def, create_temp_dir)
    test_loss = test_method(
        test_data, model_def, create_temp_dir)

    assert train_loss == test_loss


@pytest.mark.usefixtures("create_temp_dir")
@pytest.mark.usefixtures("ray_server")
@pytest.mark.parametrize(
    "model_gen", regressor_funcs)
def test_train_supervised_1_ray(create_temp_dir, model_gen):
    import ray

    # Fetch data.
    (num_train, train_data) = dataset_1['train']
    (num_test, test_data) = dataset_1['test']

    model_def = model_gen(definition=True)

    def test_model(test_data, model):
        import dryml.metrics
        test_ds = dryml.data.NumpyDataset(test_data, supervised=True)

        return dryml.metrics.mean_squared_error(model, test_ds)

    @ray.remote(num_cpus=1, num_gpus=0, max_calls=1)
    def train_method(train_data, test_data, model_def, work_dir):
        # Create repo
        import dryml
        repo = dryml.Repo(directory=work_dir)

        # Create datasets
        train_ds = dryml.data.NumpyDataset(
            train_data,
            batch_size=num_train,
            supervised=True)

        # Build and fetch model
        model = model_def.build(repo=repo)

        # Create context
        with dryml.context.ContextManager(model.dry_context_requirements()):
            # Train model
            model.prep_train()
            model.train(train_ds)

            acc = test_model(test_data, model)

        # Save model to repo.
        repo.save(model)

        return acc

    @ray.remote(num_cpus=1, num_gpus=0, max_calls=1)
    def test_method(test_data, model_def, work_dir):
        import dryml
        # Create repo
        repo = dryml.Repo(directory=work_dir)

        # Fetch model based on definition
        model = dryml.utils.head(repo.get(model_def, build_missing_def=False))

        with dryml.context.ContextManager(model.dry_context_requirements()):
            acc = test_model(test_data, model)

        return acc

    train_loss = ray.get(train_method.remote(
        train_data, test_data, model_def, create_temp_dir))
    test_loss = ray.get(test_method.remote(
        test_data, model_def, create_temp_dir))

    assert train_loss == test_loss


def gen_dataset_2():
    num_classes = 10
    num_dims = 2

    # Generate classes
    L = -10.
    H = 10.
    Max_W = 5.
    centers = (np.random.random((num_classes, num_dims))*(H-L))+L
    widths = np.random.random((num_classes, num_dims))*Max_W

    num_train = 5000
    train_data = mc.gen_dataset_2(
        num_examples=num_train,
        centers=centers,
        widths=widths)

    num_test = 100
    test_data = mc.gen_dataset_2(
        num_examples=num_test,
        centers=centers,
        widths=widths)

    return {
        'num_classes': num_classes,
        'num_dims': num_dims,
        'train': (num_train, train_data),
        'test': (num_test, test_data)
    }


dataset_2 = gen_dataset_2()


@pytest.mark.parametrize(
    "model_gen", classifier_funcs)
def test_train_supervised_2_context(model_gen):
    # Fetch data
    num_classes = dataset_2['num_classes']
    num_dims = dataset_2['num_dims']
    (num_train, train_data) = dataset_2['train']
    (num_test, test_data) = dataset_2['test']

    # generate the model
    model = model_gen(num_dims, num_classes, definition=False)

    def test_model(test_data, model):
        import dryml.metrics
        test_ds = dryml.data.NumpyDataset(test_data, supervised=True)

        return dryml.metrics.categorical_accuracy(model, test_ds)

    @dryml.compute_context(
        ctx_use_existing_context=False,
        ctx_update_objs=True)
    def train_method(train_data, test_data, model):
        # Create datasets
        train_ds = dryml.data.NumpyDataset(
            train_data,
            batch_size=num_train,
            supervised=True)

        # Train model
        model.prep_train()
        model.train(train_ds)

        acc = test_model(test_data, model)

        return acc

    @dryml.compute_context(ctx_use_existing_context=False)
    def test_method(test_data, model):
        return test_model(test_data, model)

    random_guessing_accuracy = 1./num_classes

    train_accuracy = train_method(train_data, test_data, model)
    test_accuracy = test_method(test_data, model)

    assert train_accuracy == test_accuracy
    assert train_accuracy > 3*random_guessing_accuracy


@pytest.mark.usefixtures("create_temp_dir")
@pytest.mark.parametrize(
    "model_gen", classifier_funcs)
def test_train_supervised_2_context_repo(create_temp_dir, model_gen):
    # Fetch data
    num_classes = dataset_2['num_classes']
    num_dims = dataset_2['num_dims']
    (num_train, train_data) = dataset_2['train']
    (num_test, test_data) = dataset_2['test']

    model_def = model_gen(num_dims, num_classes, definition=True)

    def test_model(test_data, model):
        import dryml.metrics
        test_ds = dryml.data.NumpyDataset(test_data, supervised=True)

        return dryml.metrics.categorical_accuracy(model, test_ds)

    model_reqs = model_def.build().dry_context_requirements()

    @dryml.compute_context(
        ctx_context_reqs=model_reqs,
        ctx_use_existing_context=False,
        ctx_update_objs=True)
    def train_method(train_data, test_data, model_def, work_dir):
        # Create repo
        repo = dryml.Repo(directory=work_dir)

        # Build model from definition
        model = model_def.build(repo=repo)

        # Create datasets
        train_ds = dryml.data.NumpyDataset(
            train_data,
            batch_size=num_train,
            supervised=True)

        # Train model
        model.prep_train()
        model.train(train_ds)

        # Save model
        repo.save(model)

        return test_model(test_data, model)

    @dryml.compute_context(
        ctx_context_reqs=model_reqs,
        ctx_use_existing_context=False)
    def test_method(test_data, model_def, work_dir):
        # Create repo
        repo = dryml.Repo(directory=work_dir)

        # Get model from the repo
        model = repo.get(model_def, build_missing_def=False)

        # Run test
        return test_model(test_data, model)

    random_guessing_accuracy = 1./num_classes

    train_accuracy = train_method(
        train_data, test_data, model_def, create_temp_dir)
    test_accuracy = test_method(
        test_data, model_def, create_temp_dir)

    assert train_accuracy == test_accuracy
    assert train_accuracy > 3*random_guessing_accuracy


@pytest.mark.usefixtures("create_temp_dir")
@pytest.mark.usefixtures("ray_server")
@pytest.mark.parametrize(
    "model_gen", classifier_funcs)
def test_train_supervised_2_ray(create_temp_dir, model_gen):
    import ray

    # Fetch data
    num_classes = dataset_2['num_classes']
    num_dims = dataset_2['num_dims']
    (num_train, train_data) = dataset_2['train']
    (num_test, test_data) = dataset_2['test']

    model_def = model_gen(num_dims, num_classes, definition=True)

    def test_model(test_data, model):
        import dryml.metrics
        test_ds = dryml.data.NumpyDataset(test_data, supervised=True)

        return dryml.metrics.categorical_accuracy(model, test_ds)

    @ray.remote(num_cpus=1, num_gpus=0, max_calls=1)
    def train_method(train_data, test_data, model_def, work_dir):
        # Create repo
        repo = dryml.Repo(directory=work_dir)

        # Build model from definition
        model = model_def.build(repo=repo)

        # Create datasets
        train_ds = dryml.data.NumpyDataset(
            train_data,
            batch_size=num_train,
            supervised=True)

        with dryml.context.ContextManager(model.dry_context_requirements()):
            # Train model
            model.prep_train()
            model.train(train_ds)

            acc = test_model(test_data, model)

        # Save model
        repo.save(model)

        return acc

    @ray.remote(num_cpus=1, num_gpus=0, max_calls=0)
    def test_method(test_data, model_def, work_dir):
        # Create repo
        repo = dryml.Repo(directory=work_dir)

        # Get model from the repo
        model = repo.get(model_def, build_missing_def=False)

        # Run test
        with dryml.context.ContextManager(model.dry_context_requirements()):
            acc = test_model(test_data, model)

        return acc

    random_guessing_accuracy = 1./num_classes

    train_accuracy = ray.get(train_method.remote(
        train_data, test_data, model_def, create_temp_dir))
    test_accuracy = ray.get(test_method.remote(
        test_data, model_def, create_temp_dir))

    assert train_accuracy == test_accuracy
    assert train_accuracy > 3*random_guessing_accuracy


@pytest.mark.usefixtures("create_temp_dir")
@pytest.mark.usefixtures("create_temp_named_file")
def test_train_test_pattern_1(create_temp_dir, create_temp_named_file):
    """
    Detect a situation, when an object is created accidentally
    with a 'parent' class.
    """

    from dryml import Workshop
    import dryml.data.transforms

    # Fetch problem data
    (num_train, train_data) = dataset_2['train']
    (num_test, test_data) = dataset_2['test']

    # Save dataset to temp file so we can load it in subordinate processes
    with open(create_temp_named_file, 'wb') as f:
        f.write(pickle.dumps((train_data, test_data)))

    # Create workshop
    shop = Workshop(work_dir=create_temp_dir)

    # Create flatten step
    flattener = dryml.data.transforms.Flatten()
    shop.repo.add_object(flattener)

    # Define a function to create a model definition
    def build_model_def_1(
            shop, n_neighbors=2, algorithm='auto', num_examples=-1):
        import dryml.models.sklearn
        import sklearn.neighbors

        flattener = dryml.utils.head(shop.repo.get(
            selector=dryml.ObjectDef(dryml.data.transforms.Flatten)))

        mdl_def = dryml.ObjectDef(
            dryml.models.sklearn.Model,
            sklearn.neighbors.KNeighborsClassifier,
            n_neighbors=n_neighbors,
            algorithm=algorithm
            )

        mdl_def = dryml.ObjectDef(
            dryml.models.sklearn.Trainable,
            model=mdl_def,
            train_fn=dryml.ObjectDef(
                dryml.models.sklearn.BasicTraining,
                num_examples=num_examples))

        mdl_def = dryml.ObjectDef(
            dryml.models.Pipe,
            flattener,
            mdl_def,)

        return mdl_def

    # Define a function to create a model definition
    def build_model_def_2(
            shop, n_neighbors=2, algorithm='auto', num_examples=-1):
        import dryml.models.sklearn
        import sklearn.neighbors

        flattener = dryml.utils.head(shop.repo.get(
            selector=dryml.ObjectDef(dryml.data.transforms.Flatten)))

        mdl_def = dryml.ObjectDef(
            dryml.models.sklearn.ClassifierModel,
            sklearn.neighbors.KNeighborsClassifier,
            n_neighbors=n_neighbors,
            algorithm=algorithm
            )

        mdl_def = dryml.ObjectDef(
            dryml.models.sklearn.Trainable,
            model=mdl_def,
            train_fn=dryml.ObjectDef(
                dryml.models.sklearn.BasicTraining,
                num_examples=num_examples))

        mdl_def = dryml.ObjectDef(
            dryml.models.Pipe,
            flattener,
            mdl_def,)

        return mdl_def

    # Create definition
    model_def = build_model_def_1(shop, n_neighbors=2, algorithm='ball_tree')

    # Define function to train model
    @dryml.compute_context(ctx_context_reqs={'default': {}})
    def train_model(model_def, repo_dir, data_filepath):
        # Load data from file
        with open(data_filepath, 'rb') as f:
            (train_data, test_data) = pickle.loads(f.read())
        train_ds = dryml.data.NumpyDataset(train_data, supervised=True)

        # Create repo
        shop = Workshop(work_dir=repo_dir)

        # Build the object
        model_obj = model_def.build(repo=shop.repo)

        # Train model object
        model_obj.train(train_ds)

        # Save all objects
        shop.repo.save(model_obj)

    # Train the model
    train_model(model_def, create_temp_dir, create_temp_named_file)

    # load objects
    shop.repo.load_objects_from_directory()

    # Check we stored one object in the repo
    assert dryml.utils.count(
        shop.repo.get(
            model_def,
            build_missing_def=False,
            sel_kwargs={'verbosity': 10})) == 1

    # Check we can't see the object in another way.
    model_def_2 = build_model_def_2(shop, n_neighbors=2, algorithm='ball_tree')
    try:
        assert dryml.utils.count(shop.repo.get(
            model_def_2,
            sel_kwargs={'verbosity': 10},
            build_missing_def=False)) == 0
    except KeyError:
        pass



from dryml.models.trainable import Trainable
from dryml.models.component import Component, \
    TrainFunction
from dryml.models.model import Model
from dryml.models.pipe import Pipe
from dryml.models.train_spec import TrainSpec
import dryml.models.utils as utils

__all__ = [
    Trainable,
    Component,
    TrainFunction,
    Model,
    Pipe,
    TrainSpec,
    utils,
]



import dryml
import objects
import copy


def test_def_1():
    """
    Test conditions under which a definition is concrete
    """

    obj_def = dryml.ObjectDef(objects.HelloStr)

    assert not obj_def.is_concrete()

    obj = obj_def.build()

    assert obj.definition().is_concrete()


def test_def_2():
    """
    A definition is only concrete if all of its components are concrete
    """

    obj = objects.TestClassC(
        objects.TestClassC2(10),
        B=objects.TestClassC2(20))

    obj_def = obj.definition()

    assert obj_def.is_concrete()

    obj_def_cpy = copy.copy(obj_def)
    del obj_def_cpy['dry_args'][0]['dry_kwargs']['dry_id']

    assert not obj_def_cpy.is_concrete()

    obj_def_cpy = copy.copy(obj_def)
    del obj_def_cpy['dry_kwargs']['B']['dry_kwargs']['dry_id']

    assert not obj_def_cpy.is_concrete()


def test_def_3():
    """
    A definition is only concrete if all of its components are concrete
    """

    obj = objects.TestNest(objects.HelloTrainableD(A=objects.TestNest(10)))

    assert obj.definition().is_concrete()

    obj_def = obj.definition()
    del obj_def['dry_args'][0]['dry_kwargs']['dry_id']

    assert not obj_def.is_concrete()

    obj_def = obj.definition()
    del obj_def['dry_args'][0]['dry_kwargs']['A']['dry_kwargs']['dry_id']

    assert not obj_def.is_concrete()


def test_def_4():
    """
    A case where a fully specified definition wasn't marked as concrete
    """

    obj = dryml.ObjectWrapper(objects.HelloTrainableD, obj_kwargs={'A': 5})

    assert obj.definition().is_concrete()


def test_def_5():
    """
    A case which looks at stripping id methods
    """

    obj = objects.TestNest(
        objects.TestNest2(
            A=objects.TestNest(5)))

    obj_def_manual = dryml.ObjectDef(
        objects.TestNest,
        dryml.ObjectDef(
            objects.TestNest2,
            A=dryml.ObjectDef(
                objects.TestNest,
                5)
            )
        )

    obj_def = obj.definition()

    assert obj_def != obj_def_manual
    assert 'dry_id' in obj_def.kwargs

    assert 'dry_id' not in obj_def_manual.kwargs
    obj_class_def = obj_def.get_cat_def()
    assert obj_class_def == obj_def_manual

    obj_class_def = obj_def.get_cat_def(recursive=False)
    del obj_def.kwargs['dry_id']
    assert obj_class_def == obj_def


def test_def_6():
    """
    A case which looks at stripping id methods
    """

    obj = objects.TestNest(('test', 'test'))
    obj_def = obj.definition().get_cat_def()

    assert type(obj_def.args[0]) is tuple
    assert obj_def.args[0][0] == 'test'
    assert obj_def.args[0][1] == 'test'

    obj = objects.TestNest(['test', 'test'])
    obj_def = obj.definition().get_cat_def()

    assert type(obj_def.args[0]) is list
    assert obj_def.args[0][0] == 'test'
    assert obj_def.args[0][1] == 'test'


def test_def_7():
    """
    A case which looks at nested definition
    building indifferent situations
    """

    # Create the data containing objects
    model_obj = objects.TestNest(10)
    opt_obj = objects.TestNest3(20, model=model_obj)
    loss_obj = objects.TestNest2(A='func')
    train_fn_obj = objects.TestNest3(
        optimizer=opt_obj,
        loss=loss_obj,
        epochs=10)

    trainable_obj = objects.TestNest3(
        model=model_obj,
        train_fn=train_fn_obj
    )

    obj_def = trainable_obj.definition()

    assert obj_def.kwargs['model'] is \
        obj_def.kwargs['train_fn'].kwargs['optimizer'].kwargs['model']


def test_def_8():
    """
    A case which looks at nested definition
    building indifferent situations
    """

    # Create the data containing objects
    model_obj = objects.TestNest(10)
    opt_obj = objects.TestNest3(20, model=model_obj)
    loss_obj = objects.TestNest2(A='func')
    train_fn_obj = objects.TestNest3(
        optimizer=opt_obj,
        loss=loss_obj,
        epochs=10)

    trainable_obj = objects.TestNest3(
        model=model_obj,
        train_fn=train_fn_obj
    )

    obj_def = trainable_obj.definition()

    # Building from plain definition
    trainable_obj_built = obj_def.build()

    assert trainable_obj_built['model'] is \
        trainable_obj_built['train_fn']['optimizer']['model']
    assert trainable_obj_built['model'].A == model_obj.A
    assert trainable_obj_built['train_fn']['optimizer'][0] == opt_obj[0]
    assert trainable_obj_built['train_fn']['epochs'] == train_fn_obj['epochs']
    assert trainable_obj_built['train_fn']['loss'].A == loss_obj.A

    # Building from 'class' definition
    trainable_obj_built = obj_def.get_cat_def(recursive=True).build()

    assert trainable_obj_built['model'] is \
        trainable_obj_built['train_fn']['optimizer']['model']
    assert trainable_obj_built['model'].A == model_obj.A
    assert trainable_obj_built['train_fn']['optimizer'][0] == opt_obj[0]
    assert trainable_obj_built['train_fn']['epochs'] == train_fn_obj['epochs']
    assert trainable_obj_built['train_fn']['loss'].A == loss_obj.A



import dryml


class SimpleObject(dryml.Object):
    def __init__(self, i, **kwargs):
        self.i = i

    def version(self):
        return 2

    def __eq__(self, rhs):
        return self.i == rhs.i



import os
import traceback
from dryml.object import Object, ObjectFactory, ObjectFile, \
    ObjectDef, change_object_cls, load_object, get_contained_objects
from dryml.config import MissingIdError
from dryml.selector import Selector
from dryml.utils import get_current_cls
from typing import Optional, Callable, Union, Mapping
import tqdm
from pprint import pprint


RepoKey = Union[Object, ObjectDef, dict, ObjectFile, Selector, str]


class RepoContainer(object):
    @staticmethod
    def from_filepath(filename: str, directory: Optional[str] = None):
        new_container = RepoContainer(directory=directory)
        new_container._filename = filename
        return new_container

    @staticmethod
    def from_object(obj: Object, filename: Optional[str] = None,
                    directory: Optional[str] = None):
        new_container = RepoContainer(directory=directory)
        new_container._obj = obj
        if filename is not None:
            new_container._filename = filename
        else:
            # Set default filename
            new_container._filename = \
                obj.dry_id+'.dry'
        return new_container

    def __init__(self, directory: Optional[str] = None):
        self._directory = directory
        self._filename = None
        self._obj = None

    def __str__(self):
        if self._obj is None:
            obj_part = "Unloaded"
        else:
            obj_part = f"{self._obj}"

        try:
            path_part = f"{self.filepath}"
        except Exception:
            path_part = "No current filepath"

        def_part = str(self.definition())

        return f"RepoContainer: {obj_part} {def_part} at {path_part}"

    @property
    def filepath(self):
        if self._filename is None:
            raise RuntimeError(
                "No filename indicated "
                "Can't load from disk.")

        if self._directory is None:
            return self._filename
        else:
            return os.path.join(self._directory, self._filename)

    def is_loaded(self) -> bool:
        if self._obj is None:
            return False
        else:
            return True

    def load(self, update: bool = True,
             reload: bool = False) -> bool:
        if self._obj is not None:
            # Object is already loaded
            return True

        # Get full filepath of file
        filepath = self.filepath

        # Load object at filepath
        try:
            self._obj = load_object(filepath, update=update, reload=reload)
            return True
        except Exception as e:
            print("There was an issue loading an object!")
            print(e)
            return False

    @property
    def obj(self):
        if self._obj is None:
            raise RuntimeError(
                "Container has not loaded the object yet")
        return self._obj

    def set_obj(self, obj: Object):
        self._obj = obj

    def get_obj(self, load=False):
        if load:
            if not self.load():
                raise RuntimeError("There was an issue loading the object!")
        return self.obj

    def save(self, directory: Optional[str] = None,
             fail_without_directory: bool = True,
             save_cache=None):
        if self._obj is not None:
            # Get object filepath
            filepath = self.filepath
            # Split filepath
            orig_dir, filename = os.path.split(filepath)
            # Correct directory if asked
            if directory is not None:
                new_dir = directory
            else:
                new_dir = orig_dir

            # If we don't want to have no directory, fail here.
            if fail_without_directory and new_dir == '':
                raise RuntimeError(
                    "Instructed not to save an object "
                    "without explicit directory")

            # Build final filepath
            filepath = os.path.join(new_dir, filename)
            self._obj.save_self(filepath)

    def unload(self):
        if self._obj is not None:
            del self._obj
            self._obj = None

    def delete(self):
        # Delete on-disk file if it exists
        if os.path.exists(self.filepath):
            os.remove(self.filepath)

    def set_directory(self, directory):
        self._directory = directory

    def set_filename(self, filename):
        self._filename = filename

    def get_contained_objects(self):
        if self._obj is None:
            return set()

        return get_contained_objects(self._obj)

    def definition(self):
        if self._obj is None:
            # We need to load the file from disk
            with ObjectFile(self.filepath) as f:
                return f.definition()
        else:
            return self._obj.definition()


# This type will act as a fascade for the various Object* types.
class Repo(object):
    def __init__(self, directory: Optional[str] = None, create: bool = False,
                 load_objects: bool = True, **kwargs):
        super().__init__(**kwargs)

        # A dictionary of objects
        self.obj_dict = {}

        self._save_objs_on_deletion = False

        if directory is not None:
            self.link_to_directory(directory, create=create,
                                   load_objects=load_objects)
        else:
            self.directory = None

    @property
    def save_objs_on_deletion(self):
        return self._save_objs_on_deletion

    @save_objs_on_deletion.setter
    def save_objs_on_deletion(self, val: bool):
        if val and self.directory is None:
            raise RuntimeError(
                "Give the repo a directory if you want it to "
                "save objects upon deletion")
        self._save_objs_on_deletion = val

    def __del__(self):
        if self.save_objs_on_deletion:
            self.save()

    def link_to_directory(self, directory: str, create: bool = False,
                          load_objects: bool = True):
        # Check that directory exists
        if not os.path.exists(directory):
            if create:
                os.makedirs(directory)
        self.directory = directory
        if load_objects:
            self.load_objects_from_directory()

    def __len__(self):
        return len(self.obj_dict)

    def add_obj_cont(self, cont: RepoContainer):
        obj_id = cont.definition().dry_id
        if obj_id in self.obj_dict:
            raise ValueError(
                f"Object {obj_id} already exists in the repo!")
        self.obj_dict[obj_id] = cont

    def load_objects_from_directory(self, directory: Optional[str] = None,
                                    selector: Optional[Callable] = None,
                                    verbose: bool = False):
        "Method to refresh the internal dictionary of objects."
        # Handle directory
        if directory is None:
            if self.directory is None:
                raise RuntimeError(
                    "No default directory selected for this repo!")
            directory = self.directory

        files = os.listdir(directory)

        num_loaded = 0
        for filename in files:
            try:
                # Load container object
                obj_cont = RepoContainer.from_filepath(
                    filename, directory=directory
                )
                # Run selector
                if selector is not None:
                    if not selector(obj_cont.definition()):
                        continue
                if obj_cont.definition().dry_id not in self.obj_dict:
                    # Add the object
                    self.add_obj_cont(obj_cont)
                    num_loaded += 1
            except Exception as e:
                print(f"WARNING! Malformed file found! {obj_cont.filepath} "
                      f"skipping load. Error was: {e}")
                if verbose:
                    print(traceback.format_exc())
        if verbose:
            print(f"Loaded {num_loaded} objects")

    def add_object(self, obj: Object, filepath: Optional[str] = None,
                   add_nested: bool = True):
        # Add a single object
        if filepath is not None:
            directory, filename = os.path.split(filepath)
            if directory == '':
                if self.directory is not None:
                    directory = self.directory
                else:
                    directory = None
        else:
            filename = None
            directory = self.directory

        # Add other nested objects into the repository too
        if add_nested:
            if hasattr(obj, '__dry_obj_container_list__'):
                for o in obj.__dry_obj_container_list__:
                    if o not in self:
                        self.add_object(o, add_nested=add_nested)

        obj_cont = RepoContainer.from_object(
            obj, directory=directory, filename=filename)
        self.add_obj_cont(obj_cont)

    def add_objects(self, obj_factory: ObjectFactory, num=1):
        # Create numerous objects from a factory function
        for i in range(num):
            obj = obj_factory(repo=self)
            self.add_object(obj)

    def make_container_handler(self,
                               load_objects: bool = True,
                               open_container: bool = True,
                               update: bool = True):

        def container_opener(obj_cont, open_container: bool = True):
            if open_container:
                return obj_cont.obj
            else:
                return obj_cont

        def container_handler(obj_cont):
            if not obj_cont.is_loaded():
                if load_objects:
                    obj_cont.load(update=update)

            return container_opener(obj_cont, open_container=open_container)

        return container_handler

    def make_filter_func(
            self,
            selector: Optional[Union[Callable, ObjectDef,
                                     Mapping, Object]] = None,
            sel_args=None, sel_kwargs=None,
            only_loaded: bool = False):
        if sel_args is None:
            sel_args = []
        if sel_kwargs is None:
            sel_kwargs = {}

        if type(selector) is ObjectDef:
            # Wrap automatically for convenience
            selector = Selector.build(selector)

        def filter_func(obj_cont):
            if only_loaded:
                if not obj_cont.is_loaded():
                    return False

            if selector is not None:
                if selector(obj_cont.definition(), *sel_args, **sel_kwargs):
                    return True
                else:
                    return False
            else:
                return True

        return filter_func

    def get_obj(
            self,
            obj_def: ObjectDef,
            load=False):
        ind_id = obj_def.dry_id
        return self.get_obj_by_id(ind_id, load=load)

    def get_obj_by_id(
            self,
            obj_id,
            load=True):
        obj_container = self.obj_dict[obj_id]
        if load:
            if not obj_container.load():
                raise RuntimeError("There was an issue loading the object!")
        return obj_container.obj

    def __contains__(
            self, item: Union[Object, ObjectDef, dict, ObjectFile]):
        if issubclass(type(item), Object) or \
           type(item) is ObjectFile:
            obj_id = item.definition().dry_id
        elif issubclass(type(item), ObjectDef):
            obj_id = item.dry_id
        elif issubclass(type(item), dict):
            obj_id = ObjectDef.from_dict(item).dry_id
        else:
            raise TypeError(
                f"Unsupported type {type(item)} for repo.contains!")
        return obj_id in self.obj_dict

    def __getitem__(
            self, key: Union[RepoKey, list, tuple, Callable]):
        """
        Easy access to objects within.

        if unpack is true, plain objects are returned
        """
        return self.get(key, open_container=True)

    def get(self,
            selector: Optional[Union[RepoKey, list, tuple, Callable]] = None,
            sel_args=None, sel_kwargs=None,
            load_objects: bool = True,
            only_loaded: bool = False,
            update: bool = True,
            open_container: bool = True,
            verbose: bool = True,
            build_missing_def=False):

        # First, handle all cases where the selector refers to a specific
        # object
        obj_id = None
        if selector is not None and \
                (isinstance(selector, Object) or
                 isinstance(selector, ObjectFile)):
            obj_id = selector.definition().dry_id
        elif selector is not None and \
                isinstance(selector, ObjectDef):
            try:
                obj_id = selector.dry_id
            except MissingIdError:
                pass
        elif selector is not None and \
                isinstance(selector, Selector):
            pass
        elif selector is not None and \
                issubclass(type(selector), dict):
            try:
                obj_id = ObjectDef.from_dict(selector).dry_id
            except MissingIdError:
                pass
        elif selector is not None and \
                type(selector) is str:
            obj_id = selector

        # Build container handler
        container_handler = \
            self.make_container_handler(
                update=update,
                load_objects=load_objects,
                open_container=open_container)

        def def_builder(sel):
            new_obj = selector.build(repo=self)
            self.add_object(new_obj)
            return new_obj

        if obj_id is not None:
            # We have a single object request
            if obj_id not in self.obj_dict:
                if type(selector) is ObjectDef and build_missing_def:
                    return def_builder(selector)
                raise KeyError(
                    f"Object {selector} (type: {type(selector)}) "
                    f"(dry_id: {obj_id}) not in the Repository.")

            obj_cont = self.obj_dict[obj_id]
            if only_loaded:
                if not obj_cont.is_loaded():
                    # No objects were found
                    return []
            result = container_handler(obj_cont)
            return result

        # Now handle all cases where we have collections of keys to query

        if type(selector) is list or \
           type(selector) is tuple:

            # We need to combine objects from each key.
            result_set = set()
            for sub_key in selector:
                try:
                    res = self.get(
                        sub_key,
                        sel_args=sel_args,
                        sel_kwargs=sel_kwargs,
                        load_objects=load_objects,
                        only_loaded=only_loaded,
                        update=update,
                        open_container=open_container,
                        verbose=verbose)
                except KeyError:
                    # Skip element
                    continue

                if res is None:
                    # Skip it if None
                    continue
                if type(res) is list:
                    # Multiple results from this key
                    for el in res:
                        result_set.add(el)
                else:
                    # A single key
                    result_set.add(el)

            results = list(result_set)
            if len(results) == 0:
                # handle the case of an empty list
                raise KeyError(
                    f"Key Collection {selector} didn't match any object!")
            elif len(results) == 1:
                # Handle a single result
                return results[0]
            return list(result_set)

        # Now we do the 'vector' methods

        # Filter the internal object list
        filter_func = \
            self.make_filter_func(
                selector,
                sel_args=sel_args,
                sel_kwargs=sel_kwargs,
                only_loaded=only_loaded)
        obj_list = list(filter(filter_func, self.obj_dict.values()))

        results = list(map(container_handler, obj_list))
        if len(results) == 0:
            if isinstance(selector, ObjectDef) and build_missing_def:
                return def_builder(selector)
            raise KeyError(f"Key {selector} didn't match any object!")
        elif len(results) == 1:
            return results[0]
        else:
            return results

    def apply(self,
              func, func_args=None, func_kwargs=None,
              selector: Optional[Callable] = None,
              sel_args=None, sel_kwargs=None,
              verbose: bool = False,
              **kwargs):
        """
        Apply a function to all objects tracked by the repo.
        We can also use a Selector to apply only to specific models
        **kwargs is passed to self.get
        """
        if func_args is None:
            func_args = []
        if func_kwargs is None:
            func_kwargs = {}

        # Create apply function
        def apply_func(obj):
            return func(obj, *func_args, **func_kwargs)

        # Get object list
        objs = self.get(
            selector=selector,
            sel_args=sel_args, sel_kwargs=sel_kwargs,
            **kwargs)

        if type(objs) is list:
            # apply function to objects
            if verbose:
                return list(map(apply_func, tqdm(objs)))
            else:
                return list(map(apply_func, objs))
        else:
            return apply_func(objs)

    def reload_objs(
            self,
            selector: Optional[Callable] = None,
            sel_args=None, sel_kwargs=None,
            update: bool = False,
            reload: bool = False,
            as_cls=None):
        if self.directory is None:
            raise RuntimeError("Repo directory needs to be set for reloading.")

        def reload_func(obj_cont):
            if not obj_cont.is_loaded():
                raise RuntimeError("Can only reload already loaded Objects")

            # Get the object
            obj = obj_cont.obj

            # Get current definition of class
            if as_cls is not None:
                cls = as_cls
            else:
                cls = type(obj)
            new_cls = get_current_cls(cls, reload=reload)

            # Set object
            obj_cont.set_obj(change_object_cls(obj, new_cls, update=True))

            # Remove old object
            del obj

        self.apply(
            reload_func,
            selector=selector, sel_args=sel_args, sel_kwargs=sel_kwargs,
            open_container=False, only_loaded=True, load_objects=False)

    def save(self,
             selector: Optional[Union[RepoKey, list, tuple, Callable]] = None,
             sel_args=None, sel_kwargs=None,
             directory: Optional[str] = None,
             recursive=True,
             error_on_none=False):

        """
        Saves the object or objects matching the input selector to disk.
        if a single object is passed and not already in the repo, it's added
        then saved.

        error_on_none: Whether to throw the KeyError, when no object is in
            the repo matching the key.
        """

        save_cache = set()

        def save_func(obj_or_cont):
            if type(obj_or_cont) is Object:
                # we have a plain dry object

                if obj_or_cont in save_cache:
                    # don't need to save, it's already done.
                    return

                if recursive:
                    contained_objs = get_contained_objects(obj_or_cont)

                    for obj in contained_objs:
                        if obj in self:
                            sub_obj_cont = self.get(obj, open_container=False)
                            save_func(sub_obj_cont)
                        else:
                            save_func(obj)

                # Save
                save_path = os.path.join(
                    directory, f"{obj_or_cont.dry_id}.dry")
                obj_or_cont.save_self(save_path)

                save_cache.add(obj_or_cont)

            else:
                # We have an object container.
                if not obj_or_cont.is_loaded():
                    raise RuntimeError(
                        "Can only save currently loaded Object")

                if obj_or_cont.obj in save_cache:
                    # don't need to save, it's already done.
                    return

                # Get/save contained objects
                if recursive:
                    contained_objs = obj_or_cont.get_contained_objects()

                    for obj in contained_objs:
                        if obj in self:
                            sub_obj_cont = self.get(obj, open_container=False)
                            save_func(sub_obj_cont)
                        else:
                            obj.save_self(os.path.join(
                                directory, f"{obj.dry_id}.dry"))

                # Save object
                obj_or_cont.save(directory=directory, save_cache=save_cache)
                save_cache.add(obj_or_cont.obj)

        # If we haven't added the object to the repo yet, add it now.
        if issubclass(type(selector), Object):
            if selector not in self:
                self.add_object(selector)

        try:
            self.apply(
                save_func,
                selector=selector, sel_args=sel_args, sel_kwargs=sel_kwargs,
                open_container=False, only_loaded=True, load_objects=False)
        except KeyError as e:
            if error_on_none:
                raise e

    def save_by_id(self,
                   obj_id, directory: Optional[str] = None):
        if directory is None:
            directory = self.directory
        self.obj_dict[obj_id].save(directory=directory)

    def save_and_cache(
            self,
            selector: Optional[Callable] = None,
            sel_args=None, sel_kwargs=None):
        "Save and then delete objects. Replace their entries with strings"

        def save_func(obj_cont):
            if not obj_cont.is_loaded():
                raise RuntimeError("Can only save currently loaded Object")

            # Save object
            obj_cont.save()
            obj_cont.unload()

        self.apply(
            save_func,
            selector=selector, sel_args=sel_args, sel_kwargs=sel_kwargs,
            open_container=False, only_loaded=True, load_objects=False)

    def update(self,
               obj: Object):
        """
        Update existing object entry with new object.
        if object doesn't exist, create entry for it.
        """

        obj_id = obj.definition().dry_id

        if obj_id not in self.obj_dict:
            # Add the object as it doesn't exist yet.
            self.add_object(obj)
        else:
            # Set the object for existing container.
            self.obj_dict[obj_id].set_obj(obj)

        # Save object to disk.
        self.save_by_id(obj_id)

    def unload(self,
               selector: Optional[Callable] = None,
               sel_args=None, sel_kwargs=None):

        def unload_func(obj_cont):
            if obj_cont.is_loaded():
                obj_cont.unload()

        self.apply(
            unload_func,
            selector=selector, sel_args=sel_args, sel_kwargs=sel_kwargs,
            open_container=False, only_loaded=True, load_objects=False)

    def delete(
            self,
            selector: Optional[Callable] = None,
            sel_args=None, sel_kwargs=None,
            only_loaded: bool = True):
        "Unload and delete from disk selected models"

        # Get all selected objects
        obj_containers = self.get(
            selector=selector, sel_args=sel_args, sel_kwargs=sel_kwargs,
            open_container=False, only_loaded=only_loaded, load_objects=False)

        def del_cont(obj_cont):
            # Delete object from repo object tracker
            obj_id = obj_cont.definition().dry_id
            del self.obj_dict[obj_id]

            # Delete object from disk
            obj_cont.delete()

            # Delete object container
            del obj_cont

        if type(obj_containers) is list:
            for obj_cont in obj_containers:
                del_cont(obj_cont)
        else:
            del_cont(obj_containers)

    def list_unique_objs(
            self,
            selector: Optional[Callable] = None,
            sel_args=None, sel_kwargs=None,
            only_loaded=False):
        "List unique object definitions yielded by a selector"
        obj_containers = self.get(
            selector=selector, sel_args=sel_args, sel_kwargs=sel_kwargs,
            open_container=False, load_objects=False)

        from dryml.utils import count
        if count(obj_containers) == 1:
            obj_containers = [obj_containers]

        results = {}

        for obj_cont in obj_containers:
            if only_loaded:
                # Skip unloaded objects
                if not obj_cont.is_loaded():
                    continue
            obj_def = obj_cont.definition()
            obj_cat_def = obj_def.get_cat_def()
            cat_id = obj_cat_def.get_category_id()

            if only_loaded:
                entry = results.get(
                    cat_id,
                    {'def': obj_cat_def, 'num': 0})
                entry['num'] += 1
                results[cat_id] = entry
            else:
                entry = results.get(
                    cat_id,
                    {'def': obj_cat_def, 'loaded': 0, 'unloaded': 0})
                if obj_cont.is_loaded():
                    entry['loaded'] += 1
                else:
                    entry['unloaded'] += 1
                results[cat_id] = entry

        for cat_id in results:
            entry = results[cat_id]
            pprint(entry['def'])
            if only_loaded:
                print(f"{entry['num']} loaded")
            else:
                print(f"{entry['loaded']}/"
                      f"{entry['loaded']+entry['unloaded']} loaded")



import pickle
from typing import List


class TrainStateException(Exception):
    pass


class TrainSpec(object):
    def __init__(self, level_steps: List[int] = None, global_steps=0):
        if level_steps is None:
            self.level_steps = [0]
        else:
            self.level_steps = level_steps
        self.global_steps = global_steps
        self.cur_level = 0

    def is_state_current(self):
        """
        Check spec state for preparedness to modify
        """
        return self.cur_level == len(self.level_steps)-1

    def descend(self):
        if self.is_state_current():
            self.level_steps.append(0)
            self.global_steps += 1
        self.cur_level += 1

    def elevate(self):
        if not self.is_state_current():
            raise TrainStateException()
        self.level_steps.pop()
        self.cur_level -= 1

    def advance(self):
        if not self.is_state_current():
            raise TrainStateException()
        self.global_steps += 1
        self.level_steps[self.cur_level] += 1

    def level_step(self):
        return self.level_steps[self.cur_level]

    def global_step(self):
        return self.global_steps

    def __str__(self):
        return f"{self.level_steps}, {self.global_steps} ({self.cur_level})"

    def __repr__(self):
        return f"{self.level_steps}, {self.global_steps} ({self.cur_level})"

    def save(self, path):
        with open(path, 'wb') as f:
            f.write(pickle.dumps({
                'level_steps': self.level_steps,
                'global_steps': self.global_steps}))

    @staticmethod
    def load(path):
        with open(path, 'rb') as f:
            def_dict = pickle.loads(f.read())
        return TrainSpec(
            level_steps=def_dict['level_steps'],
            global_steps=def_dict['global_steps'])



from dryml.utils import get_class_str, get_class_from_str, \
    apply_func
from dryml import Selector
import objects


def test_class_utils_1():
    cls = objects.HelloInt

    class_str = get_class_str(cls)

    cls_2 = get_class_from_str(class_str)

    assert cls is cls_2


def test_class_utils_2():
    obj = objects.HelloInt(msg=5)

    class_str = get_class_str(obj)

    cls_2 = get_class_from_str(class_str)

    assert type(obj) is cls_2


def test_apply_func_1():
    obj1 = objects.TestNest(objects.TestClassF1())

    def f(o):
        o.val = 20
    apply_func(obj1, f, sel=Selector(objects.TestClassF1))

    assert obj1.A.val == 20


def test_apply_func_2():
    obj1 = objects.TestClassC(
        objects.TestClassF1(),
        B=objects.TestClassC(
            objects.TestClassF1(),
            B=objects.TestClassF1()))

    def f(o):
        o.val = 20

    apply_func(obj1, f, sel=Selector(objects.TestClassF1))

    assert obj1.A.val == 20
    assert obj1.B.A.val == 20
    assert obj1.B.B.val == 20
    assert not hasattr(obj1, 'val')
    assert not hasattr(obj1.B, 'val')



"""
Utility functions for data methods
"""


def nested_flatten(data):
    flatten_data = []

    def _nested_flatten(data):
        if type(data) is dict:
            for key in data:
                _nested_flatten(data[key])
        elif type(data) is tuple:
            for el in data:
                _nested_flatten(el)
        else:
            flatten_data.append(data)

    _nested_flatten(data)
    return flatten_data


def renest_flat(shape_data, flat_data):
    def _renester(data):
        if type(data) is dict:
            res = {}
            for key in data:
                res[key] = _renester(data[key])
            return res
        elif type(data) is tuple:
            res = []
            for el in data:
                res.append(_renester(el))
            return tuple(res)
        else:
            return flat_data.pop(0)

    res = _renester(shape_data)

    return res


def nested_apply(data, func_lambda, *func_args, **func_kwargs):
    flattened_data = nested_flatten(data)
    flattened_data = list(map(
        lambda el: func_lambda(el, *func_args, **func_kwargs),
        flattened_data))
    return renest_flat(data, flattened_data)


def nestize(f, *func_args, **func_kwargs):
    return lambda el: nested_apply(el, f, *func_args, **func_kwargs)


def nested_slice(data, slicer):
    def _slicer(el):
        return el[slicer]
    return nested_apply(data, _slicer)


def get_data_batch_size(full_data=None, flat_data=None):
    if full_data is None and flat_data is None:
        raise ValueError(
            "At least one of full_data or flat_data needs not be None.")
    if full_data is not None and flat_data is not None:
        raise ValueError("Can't specify both full_data and flat_data.")
    if full_data is not None:
        flat_data = nested_flatten(full_data)
    lengths = set(map(lambda el: len(el), flat_data))
    if len(lengths) > 1:
        raise ValueError(f"Inconsistent element sizes: {lengths}")
    return lengths.pop()


def nested_batcher(data_gen, batch_size, stack_method, drop_remainder=True):
    it = iter(data_gen())
    while True:
        flat_batch_data = None
        flat_batch_shape = None
        num_collected = 0
        try:
            # Fill up batches
            while True:
                el = next(it)
                el_flat = nested_flatten(el)
                if flat_batch_data is None:
                    flat_batch_shape = el
                    flat_batch_data = list(
                        map(lambda e: list(),
                            el_flat))
                for i in range(len(el_flat)):
                    flat_batch_data[i].append(el_flat[i])
                num_collected += 1
                if num_collected >= batch_size:
                    break
        except StopIteration:
            # Catch stop iteration for partial batches
            pass
        if drop_remainder and num_collected != batch_size:
            # Exit now and don't yield
            break
        # if we have a non-empty batch, yield it.
        if flat_batch_data is not None and \
                len(flat_batch_data) > 0:
            flat_batch_data = list(map(
                stack_method,
                flat_batch_data))
            yield renest_flat(
                flat_batch_shape,
                flat_batch_data)
        else:
            break


def nested_unbatcher(data_gen):
    it = iter(data_gen())
    while True:
        try:
            d = next(it)
        except StopIteration:
            return
        flat_d = nested_flatten(d)
        length = get_data_batch_size(flat_data=flat_d)
        for i in range(length):
            new_d = list(map(lambda el: el[i], flat_d))
            yield renest_flat(d, new_d)


def taker(gen_func, n):
    i = 0
    it = iter(gen_func())
    while i < n:
        try:
            yield next(it)
            i += 1
        except StopIteration:
            return
    return


def skiper(gen_func, n):
    i = 0
    it = iter(gen_func())
    while i < n:
        try:
            next(it)
            i += 1
        except StopIteration:
            return
    while True:
        try:
            yield next(it)
        except StopIteration:
            return



import pandas as pd
import numpy as np


def prep_df(DF, feature_list, index=True):
    """
    Returns a numpy of shape (ex, features) from a pandas dataframe
    """

    # Select features we're interested in.
    sel_DF = DF[feature_list].dropna()

    if index:
        return sel_DF.to_numpy(), sel_DF.index
    else:
        return sel_DF.to_numpy()


def prep_df_lags(DF, feature_list, lags, index=True):
    """
    Returns a numpy of shape (ex, lags, features) from a pandas dataframe
    """

    # Build lag list if lags were specified as an integer
    if type(lags) is int:
        lags = list(range(lags+1))[1:]

    # Stack lagged slices
    res_npy = []
    for feature in feature_list:
        lagged_features = []
        for lag in lags:
            lagged_features.append(DF[feature].shift(lag))
        lag_features = pd.concat(lagged_features, axis=1).dropna()
        res_npy.append(lag_features)

    # Build a common index
    common_idx = res_npy[0].index
    for feat in res_npy:
        common_idx = common_idx.intersection(feat.index)

    # Normalize the index
    for i in range(len(res_npy)):
        res_npy[i] = res_npy[i].loc[common_idx].to_numpy()

    # Build result array
    res_npy = np.stack(res_npy, axis=1)

    # Return dataset and possibly index
    if index:
        return res_npy, common_idx
    else:
        return res_npy



class SaveCache(object):
    def __init__(self):
        self.save_object_cache = {}
        self.save_compute_cache = set()

    @property
    def obj_cache(self):
        return self.save_object_cache

    @property
    def compute_cache(self):
        return self.save_compute_cache

    def __del__(self):
        # Close int_files in save_cache
        for key in self.save_object_cache:
            self.save_object_cache[key].close()

    def __repr__(self):
        return f"object_cache: {self.save_object_cache} " \
           f"compute_cache: {self.save_compute_cache}"



from dryml.object import Object
from dryml.data.dataset import Dataset
from dryml.context import cls_method_compute


@cls_method_compute('eval')
class Model(Object):
    def eval(self, X: Dataset, *args, **kwargs):
        raise NotImplementedError()



from dryml.data.dataset import Dataset, NotIndexedError, \
    NotSupervisedError
from dryml.data.numpy_dataset import NumpyDataset
import dryml.data.util as util
import dryml.data.transforms as transforms


__all__ = [
    Dataset,
    NotIndexedError,
    NotSupervisedError,
    NumpyDataset,
    util,
    transforms,
]



from dryml.config import Meta
from dryml.object import Object
from dryml.data import Dataset
from dryml.models.trainable import Trainable as BaseTrainable
from dryml.models.torch.base import Wrapper
from dryml.models.torch.base import Model as TorchModel
from dryml.models.torch.base import TrainFunction as TorchTrainFunction
from dryml.models.torch.base import Trainable as TorchTrainable
from dryml.context import context
import zipfile
import torch
import tqdm


class Model(TorchModel):
    def __call__(self, *args, **kwargs):
        return self.mdl.forward(*args, **kwargs)

    def compute_cleanup_imp(self):
        del self.mdl
        self.mdl = None

    def load_compute_imp(self, file: zipfile.ZipFile) -> bool:
        try:
            with file.open('state.pth', 'r') as f:
                self.mdl.load_state_dict(torch.load(f))
            return True
        except Exception:
            return False

    def save_compute_imp(self, file: zipfile.ZipFile) -> bool:
        try:
            with file.open('state.pth', 'w') as f:
                torch.save(self.mdl.state_dict(), f)
            return True
        except Exception:
            return False

    def prep_eval(self):
        devs = context().get_torch_devices()
        self.mdl.to(devs[0])
        self.mdl.train(False)

    def prep_train(self):
        devs = context().get_torch_devices()
        self.mdl.to(devs[0])
        self.mdl.train(True)


class TrainFunction(TorchTrainFunction):
    pass


class ModelWrapper(Model):
    @Meta.collect_args
    @Meta.collect_kwargs
    def __init__(self, cls, *args, **kwargs):
        self.cls = cls
        self.args = args
        self.kwargs = kwargs
        self.mdl = None

    def compute_prepare_imp(self):
        self.mdl = self.cls(*self.args, *self.kwargs)


class Sequential(Model):
    def __init__(self, layer_defs=[]):
        self.layer_defs = layer_defs
        self.mdl = None

    def compute_prepare_imp(self):
        # create_layers
        layers = []
        for layer in self.layer_defs:
            if type(layer[0]) is not type:
                raise TypeError(
                    "First element of a layer definition should be a type")
            layers.append(layer[0](*layer[1], **layer[2]))

        self.mdl = torch.nn.Sequential(
            *layers)


class TorchOptimizer(Object):
    @Meta.collect_args
    @Meta.collect_kwargs
    def __init__(self, cls, model: Model, *args, **kwargs):
        if type(cls) is not type:
            raise TypeError("first argument must be a class!")
        self.cls = cls
        self.model = model
        self.args = args
        self.kwargs = kwargs
        self.opt = None

    def compute_prepare_imp(self):
        self.opt = self.cls(
            self.model.mdl.parameters(),
            *self.args,
            *self.kwargs)

    def load_compute_imp(self, file: zipfile.ZipFile) -> bool:
        try:
            with file.open('state.pth', 'r') as f:
                self.opt.load_state_dict(torch.load(f))
            return True
        except Exception:
            return False

    def save_compute_imp(self, file: zipfile.ZipFile) -> bool:
        try:
            with file.open('state.pth', 'w') as f:
                torch.save(self.opt.state_dict(), f)
            return True
        except Exception:
            return False

    def compute_cleanup_imp(self):
        del self.opt
        self.opt = None


class Trainable(TorchTrainable):
    def __init__(
            self,
            model: Model = None,
            train_fn: TrainFunction = None):
        self.model = model
        self.train_fn = train_fn

    def train(
            self, data, train_spec=None, train_callbacks=[],
            metrics=[]):
        self.train_fn(
            self, data, train_spec=train_spec,
            train_callbacks=train_callbacks)
        self.train_state = BaseTrainable.trained

    def prep_train(self):
        self.model.prep_train()

    def prep_eval(self):
        self.model.prep_eval()

    def eval(self, data: Dataset, *args, eval_batch_size=32, **kwargs):
        # Move variables to same device as model
        devs = context().get_torch_devices()
        if data.batched:
            # We can execute the method directly on the data
            return data.torch() \
                       .map_el(lambda el: el.to(devs[0])) \
                       .apply_X(
                           func=lambda X: self.model(X, *args, **kwargs))
        else:
            # We first need to batch the data, then unbatch to leave
            # The dataset character unchanged.
            return data.torch() \
                       .batch(batch_size=eval_batch_size) \
                       .map_el(lambda el: el.to(devs[0])) \
                       .apply_X(
                            func=lambda X: self.model(X, *args, **kwargs)) \
                       .unbatch()


class BasicTraining(TrainFunction):
    def __init__(
            self,
            optimizer: Wrapper = None,
            loss: Wrapper = None,
            epochs=1):
        self.optimizer = optimizer
        self.loss = loss
        self.epochs = epochs

    def __call__(
            self, trainable: Model, data: Dataset, train_spec=None,
            train_callbacks=[]):

        # Pop the epoch to resume from
        start_epoch = 0
        if train_spec is not None:
            start_epoch = train_spec.level_step()

        # Type checking training data, and converting if necessary
        batch_size = 32
        data = data.torch().batch(batch_size=batch_size)
        total_batches = data.count()

        # Move variables to same device as model
        devs = context().get_torch_devices()
        data = data.map_el(lambda el: el.to(devs[0]))

        # Check data is supervised.
        if not data.supervised:
            raise RuntimeError(
                f"{__class__} requires supervised data")

        optimizer = self.optimizer.opt
        loss = self.loss.obj
        model = trainable.model

        for i in range(start_epoch, self.epochs):

            running_loss = 0.
            num_batches = 0
            t_data = tqdm.tqdm(data, total=total_batches)
            for X, Y in t_data:
                optimizer.zero_grad()

                outputs = model(X)
                loss_val = loss(outputs, Y)
                loss_val.backward()
                optimizer.step()

                running_loss += loss_val.item()
                num_batches += 1
                av_loss = running_loss/(num_batches*batch_size)
                t_data.set_postfix(loss=av_loss)

            print(f"Epoch {i+1} - Average Loss: {av_loss}")
